# -*- coding: iso-8859-1 -*-
# symbolic.py
# Symbolic simulation module
# Copyright 2010-2013 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides the functionality needed to perform a small-signal symbolic
simulation.

The principal method is :func:`symbolic_analysis`, which performs the symbolic
circuit solution.

.. note::

    This module is geared towards *setting up and running* the symbolic
    simulation.  Typically, it should be used in conjunction with
    :class:`ahkab.results.symbolic_solution`, the symbolic solution class, which
    holds several convenience methods to extensively manipulate, simplify,
    post-process and analyze the simulation results, complementing the
    functionality offered by the Sympy module itself.

Reference
\"\"\"\"\"\"\"\"\"

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import sympy
from sympy.matrices import zeros as smzeros

from . import circuit
from . import devices
from . import ekv
from . import mosq
from . import diode
from . import printing
from . import py3compat
from . import results
from . import options

specs = {'symbolic': {'tokens': ({
                                  'label': 'tf',
                                  'pos': None,
                                  'type': str,
                                  'needed': False,
                                  'dest': 'source',
                                  'default': None
                                 },
                                 {
                                  'label': 'ac',
                                  'pos': None,
                                  'type': bool,
                                  'needed': False,
                                  'dest': 'ac_enable',
                                  'default': True
                                 },
                                 {
                                  'label': 'r0s',
                                  'pos': None,
                                  'type': bool,
                                  'needed': False,
                                  'dest': 'r0s',
                                  'default': False
                                 }
                                )
                     }
        }

#: the Laplace variable
s = sympy.Symbol('s', complex=True)

enabled_assumptions = {'real':True, 'positive':True, 'complex':True}

# When the MNA analysis is formulated in terms of conductances, we need
# to:
# * create a conductance symbol corresponding to the resistor symbol give by the
#   user,
# * be able to avoid collisions (distiguish) between these artificial
#   conductances and a user-defined in-circuit trans-conductance (ie a
#   VCCS or ISource), as their part_id is ``G<alphanum>``.
#
# To do so, we create symbols that are named:
# ``'G'`` + _COND_POSTFIX + part_id[1:]
#
_COND_POSTFIX = 'xxx'

def symbolic_analysis(circ, source=None, ac_enable=True, r0s=False, subs=None,
                      outfile=None, verbose=3):
    """Attempt a symbolic, small-signal solution of the circuit.

    **Parameters:**

    circ : circuit instance
        the circuit instance to be simulated.

    source : string, optional
        the ``part_id`` of the source to be used as input for the transfer
           function. If ``None``, no transfer function is evaluated.

    ac_enable : bool, optional
        take frequency dependency into consideration (default: True).

    r0s : bool, optional
        take transistors' output impedance into consideration (default: False)

    subs: dict, optional
        a dictionary of part IDs to be substituted. It makes solving the circuit
        easier. Eg. ``subs={'R1':'R2'}`` - replace the resistor R1 with R2.

    outfile : string, optional
        output filename - ``'stdout'`` means print to stdout, the default.

    verbose: int, optional
        verbosity level 0 (silent) to 6 (painful).

    **Returns:**

    sol : symbolic solution
        The solutions.

    tfs : symbolic solution
        The transfer functions, only if requested. Otherwise ``tfs`` is a
        ``None`` object.

    """
    if subs is None:
        subs = {}  # no subs by default
    else:
        subs = _parse_substitutions(subs)

    if not ac_enable:
        printing.print_info_line(
            ("Starting symbolic small-signal DC analysis...", 1), verbose)
    else:
        printing.print_info_line(
            ("Starting symbolic AC analysis...", 1), verbose)

    printing.print_info_line(
        ("Building symbolic MNA, N and x...", 3), verbose, print_nl=False)
    mna, N, subs_g = generate_mna_and_N(circ, opts={'r0s': r0s}, ac=ac_enable,
                                        subs=subs, verbose=verbose)
    x = get_variables(circ)
    mna = mna[1:, 1:]
    N = N[1:, :]
    printing.print_info_line((" done.", 3), verbose)

    printing.print_info_line(("MNA matrix (reduced):", 5), verbose)
    printing.print_info_line((sympy.sstr(mna), 5), verbose)
    printing.print_info_line(("N matrix (reduced):", 5), verbose)
    printing.print_info_line((sympy.sstr(N), 5), verbose)

    printing.print_info_line(("Building equations...", 3), verbose)
    eq = []
    for i in _to_real_list(mna * x + N):
        eq.append(sympy.Eq(i, 0))

    x = _to_real_list(x)

    if verbose > 3:
        printing.print_symbolic_equations(eq)
    printing.print_info_line(("To be solved for:", 4), verbose)
    printing.print_info_line((str(x), 4), verbose)
    printing.print_info_line(("Solving...", 1), verbose)

    sol = sympy.solve(
            eq, x, manual=options.symb_sympy_manual_solver, simplify=True)

    for ks in list(sol.keys()):
        sol.update({ks: sol[ks].subs(subs_g)})

    if sol == {}:
        printing.print_warning("No solutions. Check the netlist.")
    else:
        printing.print_info_line(("Success!", 2), verbose)
        printing.print_info_line(("Results:", 1), verbose)
        if options.cli:
            printing.print_symbolic_results(sol)

    if source is not None:
        src = _symbol_factory(source.upper())
        printing.print_info_line(("Calculating small-signal symbolic transfer functions (%s))..." %
                                 (str(src),), 2), verbose, print_nl=False)
        tfs = calculate_gains(sol, src)
        printing.print_info_line(("done.", 2), verbose)
        printing.print_info_line(
            ("Small-signal symbolic transfer functions:", 1), verbose)
        if options.cli:
            printing.print_symbolic_transfer_functions(tfs)
    else:
        tfs = None

    # convert to a results instance
    sol = results.symbolic_solution(sol, subs, circ, outfile)
    if tfs:
        if outfile and outfile != 'stdout':
            outfile += ".tfs"
        tfs = results.symbolic_solution(tfs, subs, circ, outfile, tf=True)
    return sol, tfs


def calculate_gains(sol, xin, optimize=True):
    """Calculate low-frequency gain and roots of a transfer function.

    **Parameters:**

    sol : dict
        the circuit solution
    xin : Sympy symbol
        the input variable
    optimize : boolean, optional
        If ``optimize`` is set to ``False``, no algebraic simplification
        will be attempted on the results. The default (``optimize=True``)
        results in ``sympy.together`` being called on each expression.

    **Returns:**

    gs : dict
        A dictionary with as keys the strings <key>/<xin> and as values
        dictionaries with keys ``'gain'``, ``'gain0'``, ``'poles'``,
        ``'zeros'``.

    """
    gains = {}
    for key, value in sol.items():
        tf = {}
        gain = sympy.together(value.diff(xin)) if optimize else value.diff(xin)
        (ps, zs) = get_roots(gain)
        tf.update({'gain': gain})
        tf.update({'gain0': gain.subs(s, 0)})
        tf.update({'poles': ps})
        tf.update({'zeros': zs})
        gains.update({"%s/%s" % (str(key), str(xin)): tf})
    return gains


def get_variables(circ):
    """Get a sympy matrix containing the circuit variables to be solved for.

    **Parameters:**

    circ : circuit instance
        The circuit

    **Returns:**

    vars : sympy matrix, shape (n, 1)
        The variables in a column vector.
    """
    # numero di soluzioni di tensione (al netto del ref)
    nv_1 = circ.get_nodes_number() - 1

    # descrizioni dei componenti non definibili in tensione
    idescr = [elem.part_id.upper()
              for elem in circ if circuit.is_elem_voltage_defined(elem)]

    mna_size = nv_1 + len(idescr)
    x = smzeros(mna_size, 1)

    for i in range(mna_size):
        if i < nv_1:
            x[i, 0] = _symbol_factory("V" + str(circ.nodes_dict[i + 1]))
        else:
            x[i, 0] = _symbol_factory("I[" + idescr[i - nv_1] + "]")
    return x


def _to_real_list(M):
    """
    M.tolist() returns a list of lists, even when the symb matrix is really just a vector.
    we want a list of symbols! This fixes that.

    mylist[k] = mymat.tolist[k][0]

    M: a sympy matrix with only one column

    Returns: a list.
    """
    fakelist = M.tolist()
    reallist = []
    for elem in fakelist:
        reallist.append(elem[0])
    return reallist


def generate_mna_and_N(circ, opts, ac=False, subs=None, verbose=3):
    """Generate a symbolic Modified Nodal Analysis matrix and N vector.

    Only elements that have an ``is_symbolic`` attribute set to ``True``
    (the default) are considered symbolically. Simply set the attribute to
    ``False`` to employ the numeric value. This allows to simplify and speed
    up the work of the symbolic solver.

    The formulation can be performed using conductances or resistances. The
    choice is made setting the global ``options.symb_formulate_with_gs`` value
    to ``True``. A formulation done in terms of resistors, may result in many
    separate :math:`1/R` terms in the matrices. Historically, ``sympy`` choked
    on those, because of a long-standing bug in polys. Now the issue seems to
    have been solved and the two computations should be symbolically equivalent
    albeit computationally different (as expected). The option value
    ``options.symb_formulate_with_gs`` is provided to restore the old
    functionality in case you use an old version of ``sympy``.

    **Parameters:**

    circ : circuit instance
        The circuit.
    opts : dict
        The options to be used for the generation of the matrices. As of now,
        the only supported option is ``'r0s'`` which can be set to either
        ``True`` or ``False``, and selects whether the equivalent output
        resistance of the transistors should be taken into account or not.
    ac : bool, optional
        Flag to trigger the inclusion of frequency-dependent elements. Defaults
        to ``False`` currently (but may change).
    subs : dict, optional
        The substitution dictionary, composed by mappings of
        ``<symbol>``:``<sympy expression>``.
    verbose : int, optional
        Verbosity flag, from ``0`` (silent) to ``6`` (very logorrhoic). Defaults
        to ``3``.

    **Returns:**

    mna, N : Sympy matrices
        The MNA matrix and the contant term of symbolic type.
    subs_gs : dict of symbols
        In case the formulation of the MNA is performed in terms of
        conducatances, this dictionary is to be used to substitute away the
        conducatances for the resistor symbols, after the circuit is solved but
        before the results are shown to the user. ``sympy``'s ``sub()`` can take
        care of that for you. If not necessary, this dictionary is empty.

    .. note::

        Setting ``opts['r0s'] = True``, ie considering all the transistors output
        resistances, can significantly slow down -- or even prevent by consuming
        all available memory -- the solution of complex circuits with several
        active elements.

        We recommend a combination of the following:

        * setting the above option in simple circuits only,
        * inserting explicitely the :math:`r_0` you wish to consider at circuit
          level,
        * beefing up your machine with extra RAM and extra computing power,
        * being patient.
    """
    n_of_nodes = circ.get_nodes_number()
    mna = smzeros(n_of_nodes)
    N = smzeros(n_of_nodes, 1)
    subs_g = {}

    for elem in circ:
        if isinstance(elem, devices.Resistor):
            # we use conductances instead of 1/R because there is a significant
            # overhead handling many 1/R terms in sympy.
            if elem.is_symbolic:
                R = _symbol_factory(elem.part_id.upper(), real=True,
                                    positive=True)
                if R in subs:  # instant substitution
                    R = subs[R]
                if options.symb_formulate_with_gs:
                    G = _symbol_factory('Gxxx' + str(R)[1:], real=True,
                                        positive=True)
                    # but we keep track of which is which and substitute back after
                    # solving.
                    subs_g.update({G:1/R})
                else:
                    G = 1.0/R
            else:
                    G = 1.0/elem.value
            mna[elem.n1, elem.n1] = mna[elem.n1, elem.n1] + G
            mna[elem.n1, elem.n2] = mna[elem.n1, elem.n2] - G
            mna[elem.n2, elem.n1] = mna[elem.n2, elem.n1] - G
            mna[elem.n2, elem.n2] = mna[elem.n2, elem.n2] + G
        elif isinstance(elem, devices.Capacitor):
            if ac:
                if elem.is_symbolic:
                    capa = _symbol_factory(
                        elem.part_id.upper(), real=True, positive=True)
                    if capa in subs:  # instant substitution
                        capa = subs[capa]
                else:
                    capa = elem.value
                mna[elem.n1, elem.n1] = mna[elem.n1, elem.n1] + s * capa
                mna[elem.n1, elem.n2] = mna[elem.n1, elem.n2] - s * capa
                mna[elem.n2, elem.n2] = mna[elem.n2, elem.n2] + s * capa
                mna[elem.n2, elem.n1] = mna[elem.n2, elem.n1] - s * capa
            else:
                pass
        elif isinstance(elem, devices.Inductor):
            pass
        elif isinstance(elem, devices.GISource):
            if elem.is_symbolic:
                alpha = _symbol_factory(elem.part_id.upper(), real=True)
                if alpha in subs:  # instant substitution
                    alpha = subs[alpha]
            else:
                alpha = elem.value
            mna[elem.n1, elem.sn1] = mna[elem.n1, elem.sn1] + alpha
            mna[elem.n1, elem.sn2] = mna[elem.n1, elem.sn2] - alpha
            mna[elem.n2, elem.sn1] = mna[elem.n2, elem.sn1] - alpha
            mna[elem.n2, elem.sn2] = mna[elem.n2, elem.sn2] + alpha
        elif isinstance(elem, devices.ISource):
            if elem.is_symbolic:
                IDC = _symbol_factory(elem.part_id.upper())
                if IDC in subs:  # instant substitution
                    IDC = subs[IDC]
            else:
                IDC = elem.dc_value
            N[elem.n1, 0] = N[elem.n1, 0] + IDC
            N[elem.n2, 0] = N[elem.n2, 0] - IDC
        elif isinstance(elem, mosq.mosq_device) or isinstance(elem, ekv.ekv_device):
            gm = _symbol_factory('gm_' + elem.part_id, real=True, positive=True)
            if gm in subs:  # instant substitution
                gm = subs[gm]
            mna[elem.n1, elem.ng] = mna[elem.n1, elem.ng] + gm
            mna[elem.n1, elem.n2] = mna[elem.n1, elem.n2] - gm
            mna[elem.n2, elem.ng] = mna[elem.n2, elem.ng] - gm
            mna[elem.n2, elem.n2] = mna[elem.n2, elem.n2] + gm
            if opts['r0s']:
                r0 = _symbol_factory(
                    'r0_' + elem.part_id, real=True, positive=True)
                if r0 in subs:  # instant substitution
                    r0 = subs[r0]
                mna[elem.n1, elem.n1] = mna[elem.n1, elem.n1] + 1 / r0
                mna[elem.n1, elem.n2] = mna[elem.n1, elem.n2] - 1 / r0
                mna[elem.n2, elem.n1] = mna[elem.n2, elem.n1] - 1 / r0
                mna[elem.n2, elem.n2] = mna[elem.n2, elem.n2] + 1 / r0
        elif isinstance(elem, diode.diode):
            gd = _symbol_factory("g" + elem.part_id, positive=True)
            if gd in subs:  # instant substitution
                gd = subs[gd]
            mna[elem.n1, elem.n1] = mna[elem.n1, elem.n1] + gd
            mna[elem.n1, elem.n2] = mna[elem.n1, elem.n2] - gd
            mna[elem.n2, elem.n1] = mna[elem.n2, elem.n1] - gd
            mna[elem.n2, elem.n2] = mna[elem.n2, elem.n2] + gd
        elif isinstance(elem, devices.FISource):
            # These are added after all VDEs have been accounted for
            pass
        elif isinstance(elem, devices.InductorCoupling):
            pass
            # this is taken care of within the inductors
        elif circuit.is_elem_voltage_defined(elem):
            pass
            # we'll add its lines afterwards
        elif verbose:
            printing.print_warning("Skipped elem %s: not implemented." %
                                   (elem.part_id.upper(),))

    pre_vde = mna.shape[0]
    for elem in circ:
        if circuit.is_elem_voltage_defined(elem):
            index = mna.shape[0]  # get_matrix_size(mna)[0]
            mna = _expand_matrix(mna, add_a_row=True, add_a_col=True)
            N = _expand_matrix(N, add_a_row=True, add_a_col=False)
            # KCL
            mna[elem.n1, index] = +1
            mna[elem.n2, index] = -1
            # KVL
            mna[index, elem.n1] = +1
            mna[index, elem.n2] = -1
            if isinstance(elem, devices.VSource):
                if elem.is_symbolic:
                    VDC = _symbol_factory(elem.part_id.upper())
                    if VDC in subs:  # instant substitution
                        VDC = subs[VDC]
                else:
                    VDC = elem.dc_value
                N[index, 0] = -VDC
            elif isinstance(elem, devices.EVSource):
                if elem.is_symbolic:
                    alpha = _symbol_factory(elem.part_id.upper(), real=True)
                    if alpha in subs:  # instant substitution
                        alpha = subs[alpha]
                else:
                    alpha = elem.alpha
                mna[index, elem.sn1] = -alpha
                mna[index, elem.sn2] = +alpha
            elif isinstance(elem, devices.HVSource):
                if elem.is_symbolic:
                    alpha = _symbol_factory(elem.part_id.upper(), real=True)
                    if alpha in subs:  # instant substitution
                        alpha = subs[alpha]
                else:
                    alpha = elem.alpha
                source_index = circ.find_vde_index(elem.source_id)
                mna[index, n_of_nodes + source_index] = +alpha
            elif isinstance(elem, devices.Inductor):
                if ac:
                    if elem.is_symbolic:
                        L = _symbol_factory(
                            elem.part_id.upper(), real=True, positive=True)
                        if L in subs:  # instant substitution
                            L = subs[L]
                    else:
                        L = elem.L
                    mna[index, index] = -s * L
                else:
                    pass
                    # already so: commented out
                    # N[index,0] = 0
            else:
                raise circuit.CircuitError('Element %s is not supported. ' +
                                           'Please report this bug.' %
                                           elem.__class__)

    for elem in circ:
        if ac and isinstance(elem, devices.Inductor):
            # find its index to know which column corresponds to its
            # current
            this_index = circ.find_vde_index(elem.part_id, verbose=0)
            for cd in elem.coupling_devices:
                if cd.is_symbolic:
                    M = _symbol_factory(
                        cd.part_id, real=True, positive=True)
                    if M in subs:  # instant substitution
                        M = subs[M]
                else:
                    M = cd.K
                # get `part_id` of the other inductor (eg. "L32")
                other_id_wdescr = cd.get_other_inductor(elem.part_id)
                # find its index to know which column corresponds to
                # its current
                other_index = circ.find_vde_index(
                    other_id_wdescr, verbose=0)
                # add the term.
                mna[pre_vde + this_index,
                    pre_vde + other_index] += -s * M
        elif isinstance(elem, devices.FISource):
            source_current_index = circ.find_vde_index(elem.source_id, verbose=0)
            if elem.is_symbolic:
                F = _symbol_factory(elem.part_id, real=True)
                if F in subs:  # instant substitution
                    F = subs[F]
            else:
                F = elem.alpha
            mna[elem.n1, pre_vde + source_current_index] += +F
            mna[elem.n2, pre_vde + source_current_index] += -F
        else:
            pass

    # all done
    return mna, N, subs_g


def _expand_matrix(mat, add_a_row=False, add_a_col=False):
    if add_a_row:
        row = smzeros(1, mat.shape[1])
        mat = mat.row_insert(mat.shape[0], row)
    if add_a_col:
        col = sympy.zeros(mat.shape[0], 1)
        mat = mat.col_insert(mat.shape[1], col)
    return mat


def get_roots(expr):
    """Given the transfer function ``expr``, returns ``poles, zeros``.
    """
    num, den = sympy.fraction(expr)
    return sympy.solve(den, s), sympy.solve(num, s)


def _parse_substitutions(sdict):
    """Generates a symbolic substitution dictionary from a dictionary of
    substitutions expressed in terms of part IDs.

    This method is typically called by :func:`symbolic_analysis`.

    .. note::

        For what regards the resistors, the method will check the value
        of ``options.symb_formulate_with_gs`` and convert the resistors
        to conductances as needed, according to the option value.

    **Parameters:**

    sdict : a dict of strings
        The elements of the dictionary should be according to the syntax
        ``{'<part_id1>:<part_id2>'}``, eg ``{'R2':'R1'}``, instructing the
        simulator to substitute ``R1`` in place of ``R2``.

    **Returns:**

    subs : dict
        the dictionary of symbols to be passed to :func:`symbolic_analysis`.

    """
    subs = {}
    for v1, v2 in py3compat.iteritems(sdict):
        if v1[0].upper() == 'R' and options.symb_formulate_with_gs:
            letter_id1 = 'G' + _COND_POSTFIX
            letter_id2 = 'G' + _COND_POSTFIX
        else:
            letter_id1 = v1[0].upper()
            letter_id2 = v2[0].upper()
        if letter_id1[0] in ('R', 'G', 'L', 'C', 'M'):
            s1 = _symbol_factory(letter_id1 + v1[1:], real=True, positive=True)
        else:
            s1 = _symbol_factory(letter_id1 + v1[1:], real=True)
        if letter_id2[0] in ('R', 'G', 'L', 'C', 'M'):
            s2 = _symbol_factory(letter_id2 + v2[1:], real=True, positive=True)
        else:
            s2 = _symbol_factory(letter_id2 + v2[1:], real=True)
        subs.update({s1:s2})
    return subs

def _symbol_factory(name, **options):
    filtered_options = {}
    for i in options:
        if options[i] and enabled_assumptions[i]:
            filtered_options.update({i:options[i]})
        else:
            pass # discarded
    return sympy.Symbol(name.upper(), **filtered_options)



# -*- coding: iso-8859-1 -*-
# devices.py
# Devices for simulation
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.
"""
This module contains several basic element classes.

Introduction
------------

While they may be instantiated directly by the user, notice that the
main ``ahkab`` module provides convenience functions to instantiate
and connect into a circuit instance all of the following devices.

Notice that the circuit elements are not restricted to those provided here, the
user is welcome to provide his own. Please see the dedicated section below.

Classes defined in this module
------------------------------

.. autosummary::
    ISource
    VSource
    Resistor
    Capacitor
    Inductor
    InductorCoupling
    EVSource
    GISource
    HVSource
    FISource

Defining new elements and subclassing ``Component``
---------------------------------------------------

We recommend to subclass :class:`ahkab.devices.Component` if you intend to
define a new element.

The general form of a (possibly nonlinear) element class is described in the
following.

Required attributes and methods
===============================

The class must provide:

1. Element terminals:

::

    elem.n1 # the anode of the element
    elem.n2 # the cathode of the element

.. note:: a positive current is a current that flows into the anode and out of
    the cathode. This convention is used throughout the simulator.

2. ``elem.get_ports()``

This method must return a tuple of pairs of nodes.

Eg.

::

    ((na, nb), (nc, nd), (ne, nf), ... )

Each pair of nodes is used to determine a voltage that has effect on the
current.

For example, the source-referred model of an nmos may provide:

::

    ((n_gate, n_source), (n_drain, n_source))

The positive terminal is the first.

From that, the calling method builds a voltage vector corresponding to the
ports vector:

::

    voltages_vector = ( Va-Vb, Vc-Vd, Ve-Vf, ...)

That's passed to:

3. ``elem.i(voltages_vector, time)``

It returns the current flowing into the element if the voltages specified in
the voltages_vector are applied to its ports, at the time given.

4. ``elem.g(voltages_vector, port_index, time)``

similarly returns the differential transconductance between the port at
position ``port_index`` in the ``ports_vector`` (see point **2** above)
and the element output current, when the operating point is specified by
the voltages in the ``voltages_vector``.

5. ``elem.is_nonlinear``

A non linear element must have a ``elem.is_nonlinear`` field set to True.

6. ``elem.is_symbolic``

This boolean flag is used to know whether the element should be treated
symbolically by the ymbolic solver or not. It is meant to be toggled
by the user at will.

7. Every element should have a ``get_netlist_elem_line(self, nodes_dict)``
allowing the element to print a netlist entry that parses to itself.

Recommended attributes and methods
==================================

1. A non linear element may have a list/tuple of the same length of its
``ports_vector`` in which there are the recommended guesses for DC and OP
analyses.

Eg. ``Vgs`` is set to ``Vt0`` in mosfets.

This is obviously useless for linear devices.

Module reference
----------------

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import numpy as np

class Component(object):

    """Base Component class.

    This component is not meant for direct use, rather all other (simple)
    components are a subclass of this element.

    """

    def __init__(self, part_id=None, n1=None, n2=None, is_nonlinear=False, is_symbolic=True, value=None):
        self.part_id = part_id
        self.n1 = n1
        self.n2 = n2
        self.value = value
        self.is_nonlinear = is_nonlinear
        self.is_symbolic = is_symbolic

    #   Used by `get_netlist_elem_line` for value
    def __str__(self):
        return str(self.value)

    #   must be called to define the element!
    def set_char(self, i_function=None, g_function=None):
        if i_function:
            self.i = i_function
        if g_function:
            self.g = g_function

    def g(self, v):
        return 1./self.value

    def i(self, v):
        return 0

    def get_netlist_elem_line(self, nodes_dict):
        """A netlist line that, parsed, evaluates to the same instance

        **Parameters:**

        nodes_dict : dict
            The nodes dictionary of the circuit, so that the method
            can convert its internal node IDs to the corresponding
            external ones.

        **Returns:**

        ntlst_line : string
            The netlist line.
        """
        return "%s %s %s %g" % (self.part_id, nodes_dict[self.n1],
                                nodes_dict[self.n2], self.value)


class Resistor(Component):
    """A resistor.

    .. image:: images/elem/resistor.svg

    **Parameters:**

    part_id : string
        The unique identifier of this element. The first letter should be
        ``'R'``.
    n1 : int
        *Internal* node to be connected to the anode.
    n2 : int
        *Internal* node to be connected to the cathode.
    value : float
        Resistance in ohms.

     """
    #
    #             /\    /\    /\
    #     n1 o---+  \  /  \  /  \  +---o n2
    #                \/    \/    \/
    #
    def __init__(self, part_id, n1, n2, value):
        self.part_id = part_id
        self._value = value
        self._g = 1./value
        self.is_nonlinear = False
        self.is_symbolic = True
        self.n1 = n1
        self.n2 = n2

    @property
    def g(self, v=0, time=0):
        return self._g

    @g.setter
    def g(self, g):
        self._g = g
        self._value = 1./g

    @property
    def value(self, v=0, time=0):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        self._g = 1./value

    def i(self, v, time=0):
        return 0

    def get_op_info(self, ports_v):
        """Information regarding the Operating Point (OP)

        **Parameters:**

        ports_v : list of lists
            The parameter is to be set to ``[[v]]``, where ``v`` is the voltage
            applied to the resistor terminals.

        **Returns:**

        op_keys : list of strings
            The labels corresponding to the numeric values in ``op_info``.
        op_info : list of floats
            The values corresponding to ``op_keys``.
        """
        vn1n2 = float(ports_v[0][0])
        in1n2 = float(ports_v[0][0]/self.value)
        power = float(ports_v[0][0] ** 2 / self.value)
        op_keys = ['Part ID', u"R [\u2126]", "V(n1,n2) [V]", "I(n1->n2) [A]", "P [W]"]
        op_info = [self.part_id.upper(), self.value, vn1n2, in1n2, power]
        return op_keys, op_info

class Capacitor(Component):
    """A capacitor.

    .. image:: images/elem/capacitor.svg

    **Parameters:**

    part_id : string
        The unique identifier of this element. The first letter should be
        ``'C'``.
    n1 : int
        *Internal* node to be connected to the anode.
    n2 : int
        *Internal* node to be connected to the cathode.
    value : float
        The capacitance in Farads.
    ic : float
        The initial condition (IC) to be used for time-based simulations,
        such as TRAN analyses, when requested, expressed in Volt.

    """
    #
    #               |  |
    #               |  |
    #     n1 o------+  +-------o n2
    #               |  |
    #               |  |
    #
    def __init__(self, part_id, n1, n2, value, ic=None):
        self.part_id = part_id
        self.value = value
        self.n1 = n1
        self.n2 = n2
        self.ic = ic
        self.is_nonlinear = False
        self.is_symbolic = True

    def g(self, v, time=0):
        return 0

    def i(self, v, time=0):
        return 0

    def d(self, v, time=0):
        return self.value

    def get_op_info(self, ports_v):
        """Information regarding the Operating Point (OP)

        **Parameters:**

        ports_v : list of lists
            The parameter is to be set to ``[[v]]``, where ``v`` is the voltage
            applied to the capacitor terminals.

        **Returns:**

        op_keys : list of strings
            The labels corresponding to the numeric values in ``op_info``.
        op_info : list of floats
            The values corresponding to ``op_keys``.
        """
        vn1n2 = float(ports_v[0][0])
        qn1n2 = float(ports_v[0][0] * self.value)
        energy = float(.5 * ports_v[0][0] ** 2 * self.value)
        op_keys = ['Part ID', "V(n1-n2) [V]", "Q [C]", "E [J]"]
        op_info = [self.part_id.upper(), vn1n2, qn1n2, energy]
        return op_keys, op_info

class Inductor(Component):
    """An inductor.

    .. image:: images/elem/inductor.svg

    **Parameters:**

    part_id : string
        The unique identifier of this element. The first letter should be
        ``'L'``.
    n1 : int
        *Internal* node to be connected to the anode.
    n2 : int
        *Internal* node to be connected to the cathode.
    value : float
        The inductance in Henry.
    ic : float
        The initial condition (IC) to be used for time-based simulations,
        such as TRAN analyses, when requested, expressed in Ampere.

    """
    #
    #             L
    #  n1 o----((((((((----o n2
    #
    #
    def __init__(self, part_id, n1, n2, value, ic=None):
        self.value = value
        self.n1 = n1
        self.n2 = n2
        self.part_id = part_id
        self.ic = ic
        self.coupling_devices = []
        self.is_nonlinear = False
        self.is_symbolic = True

    def get_op_info(self, ports_v, current):
        """Information regarding the Operating Point (OP)

        **Parameters:**

        ports_v : list of lists
            The parameter is to be set to ``[[v]]``, where ``v`` is the voltage
            applied to the inductor terminals.
        current : float
            The current flowing in the inductor, positive currents flow in ``n1``
            and out of ``n2``.

        **Returns:**

        op_keys : list of strings
            The labels corresponding to the numeric values in ``op_info``.
        op_info : list of floats
            The values corresponding to ``op_keys``.
        """
        vn1n2 = float(ports_v[0][0])
        energy = .5 * self.value * current**2
        op_keys = ['Part ID', u"\u03d5(n1,n2) [Wb]", "I(n1->n2) [A]", "E [J]"]
        op_info = [self.part_id.upper(), self.value*current, current, energy]
        return op_keys, op_info


class InductorCoupling(Component):
    """Coupling between two inductors.

    .. image:: images/elem/mutual_inductors.svg

    This element is used to simulate the coupling between two inductors,
    such as in the case of a transformer.

    Notice that turn ratio and the inductance ratio are linked by the
    relationship:

    .. math::

        \\frac{L_1}{L_2} = \\left(\\frac{N_1}{N_2}\\right)^2

    **Parameters:**

    part_id : string
        The unique identifier of this element. The first letter should be
        ``'K'``.
    L1 : string
        The ``part_id`` of the first inductor to be coupled.
    L2 : string
        The ``part_id`` of the second inductor to be coupled.
    K : float
        The coupling coefficient between the two windings.
    M : float
        The mutual inductance between the windings, it is equal to
        :math:`K\\sqrt(L_1L2)`, where :math:`L_1` and :math:`L_2` are the
        values of the two inductors ``L1`` and ``L2``.
    """
    # K1 L1 L2 k=<float>
    # M = sqrt(L1elem.value * L2elem.value) * Kvalue
    def __init__(self, part_id, L1, L2, K, M):
        self.part_id = part_id
        self.L1 = L1
        self.L2 = L2
        self.M = M
        self.K = K
        self.is_nonlinear = False
        self.is_symbolic = True

    def __str__(self):
        return "%s %s %g" % (self.L1, self.L2, self.value)

    def get_other_inductor(self, Lselected):
        Lret = None
        if Lselected.upper() == self.L1.upper():
            Lret = self.L2
        elif Lselected.upper() == self.L2.upper():
            Lret = self.L1
        if Lret is None:
            raise Exception("Mutual inductors bug.")
        return Lret

    def get_netlist_elem_line(self, nodes_dict):
        """A netlist line that, parsed, evaluates to the same instance

        **Parameters:**

        nodes_dict : dict
            The nodes dictionary of the circuit, so that the method
            can convert its internal node IDs to the corresponding
            external ones.

        **Returns:**

        ntlst_line : string
            The netlist line.
        """
        return "%s %s %s %g" % (self.part_id, self.L1, self.L2, self.K)


#
# SOURCES
#


class ISource(Component):
    """An ideal current source.

    .. image:: images/elem/isource.svg

    Defaults to a DC current source.

    To implement a time-varying source:

    * set ``_time_function`` to an appropriate instance having a
      ``value(self, time)`` method,
    * set ``is_timedependent`` to ``True``.

    **Parameters:**

    part_id : string
        The unique identifier of this element. The first letter should be
        ``'I'``.
    n1 : int
        *Internal* node to be connected to the anode.
    n2 : int
        *Internal* node to be connected to the cathode.
    dc_value : float
        DC voltage in Ampere.
    ac_value : complex float, optional
        AC current in Ampere. Defaults to no AC characteristics,
        ie :math:`I(\\omega) = 0 \\;\\;\\forall \\omega > 0`.

    """
    def __init__(self, part_id, n1, n2, dc_value=None, ac_value=0):
        self.part_id = part_id
        self.dc_value = dc_value
        self.abs_ac = np.abs(ac_value) if ac_value else None
        self.arg_ac = np.angle(ac_value) if ac_value else None
        self.n1 = n1
        self.n2 = n2
        self.is_nonlinear = False
        self.is_symbolic = True
        self.is_timedependent = False
        self._time_function = None

    def __str__(self):
        rep = ""
        if self.dc_value is not None:
            rep = rep + "type=idc value=" + str(self.dc_value) + " "
        if self.abs_ac is not None:
            rep = rep + "iac=" + \
                str(self.abs_ac) + " " + "arg=" + str(self.arg_ac) + " "
        if self.is_timedependent:
            rep = rep + str(self._time_function)
        return rep

    def I(self, time=None):
        """Evaluate the current forced by the current source.

        If ``time`` is not supplied, or if it is set to ``None``, or if the
        source is only specified for DC, returns ``dc_value``.

        **Parameters:**

        time : float or None, optional
            The time at which the current is evaluated, if any.

        **Returns:**

        I : float
            The current, in Ampere.

        .. note::

            This simulator uses passive convention:
            A positive currents flows in a element into the positive node and
            out of the negative node
        """
        if not self.is_timedependent or (self._time_function is None) or \
            (time is None and self.dc_value is not None):
            return self.dc_value
        else:
            return self._time_function(time)

    def get_netlist_elem_line(self, nodes_dict):
        """A netlist line that, parsed, evaluates to the same instance

        **Parameters:**

        nodes_dict : dict
            The nodes dictionary of the circuit, so that the method
            can convert its internal node IDs to the corresponding
            external ones.

        **Returns:**

        ntlst_line : string
            The netlist line.
        """
        rep = ""
        rep += "%s %s %s " % (self.part_id, nodes_dict[self.n1],
                             nodes_dict[self.n2])
        if self.dc_value is not None:
            rep = rep + "type=idc value=" + str(self.dc_value) + " "
        if self.abs_ac is not None:
            rep = rep + "iac=" + \
                str(self.abs_ac) + " " + "arg=" + str(self.arg_ac) + " "
        if self.is_timedependent:
            rep = rep + str(self._time_function)
        return rep

    def get_op_info(self, ports_v):
        """Information regarding the Operating Point (OP)

        **Parameters:**

        ports_v : list of lists
            The parameter is to be set to ``[[v]]``, where ``v`` is the voltage
            applied to the current source terminals.

        **Returns:**

        op_keys : list of strings
            The labels corresponding to the numeric values in ``op_info``.
        op_info : list of floats
            The values corresponding to ``op_keys``.
        """
        vn1n2 = float(ports_v[0][0])
        power = float(ports_v[0][0] * self.I())
        op_keys = ['Part ID', "V(n1-n2) [V]", "I [A]", "P [W]"]
        op_info = [self.part_id.upper(), vn1n2, self.I(), power]
        return op_keys, op_info

class VSource(Component):
    """An ideal voltage source.

    .. image:: images/elem/vsource.svg

    Defaults to a DC voltage source.

    To implement a time-varying source:

    * set ``_time_function`` to an appropriate instance having a
      ``value(self, time)`` method,
    * set ``is_timedependent`` to ``True``.

    **Parameters:**

    part_id : string
        The unique identifier of this element. The first letter should be
        ``'V'``.
    n1 : int
        *Internal* node to be connected to the anode.
    n2 : int
        *Internal* node to be connected to the cathode.
    dc_value : float
        DC voltage in Volt.
    ac_value : complex float, optional
        AC voltage in Volt. Defaults to no AC characteristics,
        ie :math:`V(\\omega) = 0 \\;\\;\\forall \\omega > 0`.

    """

    def __init__(self, part_id, n1, n2, dc_value, ac_value=0):
        self.part_id = part_id
        self.dc_value = dc_value
        self.n1 = n1
        self.n2 = n2
        self.abs_ac = np.abs(ac_value) if ac_value else None
        self.arg_ac = np.angle(ac_value) if ac_value else None
        self.is_nonlinear = False
        self.is_symbolic = True
        self.is_timedependent = False
        self._time_function = None
        if dc_value is not None:
            self.dc_guess = [self.dc_value]

    def __str__(self):
        rep = ""
        if self.dc_value is not None:
            rep = rep + "type=vdc value=" + str(self.dc_value) + " "
        if self.abs_ac is not None:
            #   TODO:   netlist parser doesn't accept `arg=` from `self.arg_ac`
            rep = rep + "vac=" + str(self.abs_ac) + " "
        if self.is_timedependent:
            rep = rep + str(self._time_function)
        return rep

    def V(self, time=None):
        """Evaluate the voltage applied by the voltage source.

        If ``time`` is not supplied, or if it is set to ``None``, or if the
        source is only specified for DC, returns ``dc_value``.

        **Parameters:**

        time : float or None, optional
            The time at which the voltage is evaluated, if any.

        **Returns:**

        V : float
            The voltage, in Volt.
        """

        if (not self.is_timedependent or\
            self._time_function is None) or \
                (time is None and self.dc_value is not None):
            return self.dc_value
        else:
            return self._time_function(time)

    def get_netlist_elem_line(self, nodes_dict):
        """A netlist line that, parsed, evaluates to the same instance

        **Parameters:**

        nodes_dict : dict
            The nodes dictionary of the circuit, so that the method
            can convert its internal node IDs to the corresponding
            external ones.

        **Returns:**

        ntlst_line : string
            The netlist line.
        """
        rep = ""
        rep += "%s %s %s " % (self.part_id, nodes_dict[self.n1],
                             nodes_dict[self.n2])
        if self.dc_value is not None:
            rep = rep + "type=vdc value=" + str(self.dc_value) + " "
        if self.abs_ac is not None:
            #   TODO:   netlist parser doesn't accept `arg=` from `self.arg_ac`
            rep = rep + "vac=" + str(self.abs_ac) + " "
        if self.is_timedependent:
            rep = rep + str(self._time_function)
        return rep

    def get_op_info(self, ports_v, current):
        """Information regarding the Operating Point (OP)

        **Parameters:**

        ports_v : list of lists
            The parameter is to be set to ``[[v]]``, where ``v`` is the voltage
            applied to the source terminals.
        current : float
            The current flowing in the voltage source, positive currents flow in
            ``n1`` and out of ``n2``.

        **Returns:**

        op_keys : list of strings
            The labels corresponding to the numeric values in ``op_info``.
        op_info : list of floats
            The values corresponding to ``op_keys``.
        """
        vn1n2 = float(ports_v[0][0])
        power = self.V() * current
        op_keys = ['Part ID', "V(n1,n2) [V]", "I(n1->n2) [A]", "P [W]"]
        op_info = [self.part_id.upper(), self.V(), current, power]
        return op_keys, op_info


class EVSource(Component):
    """Linear voltage-controlled voltage source

    .. image:: images/elem/vcvs.svg

    The source port is an open circuit, the destination port is an ideal
    voltage source.

    Mathematically, it is equivalent to the following:

    .. math::

        \\left\\{
        \\begin{array}{ll}
            I_s = 0\\\\
            Vn_1 - Vn_2 = \\alpha * (Vsn_1 - Vsn_2)
        \\end{array}
        \\right.

    Where :math:`I_s` is the current at the source port and the remaining
    symbols are explained in the Parameters section below.

    **Parameters:**

    n1 : int
        *Internal* node to be connected to the anode of the output port.
    n2 : int
        *Internal* node to be connected to the cathode of the output port.
    value : float
        Proportionality constant :math:`\\alpha` between the voltages.
    sn1 : int
        *Internal* node to be connected to the anode of the source (sensing)
        port.
    sn2 : int
        *Internal* node to be connected to the cathode of the source
        (sensing) port.
    """

    def __init__(self, part_id, n1, n2, value, sn1, sn2):
        self.part_id = part_id
        self.n1 = n1
        self.n2 = n2
        self.alpha = value
        self.sn1 = sn1
        self.sn2 = sn2
        self.is_nonlinear = False
        self.is_symbolic = True

    def __str__(self):
        return "alpha=%s" % self.alpha

    def get_netlist_elem_line(self, nodes_dict):
        """A netlist line that, parsed, evaluates to the same instance

        **Parameters:**

        nodes_dict : dict
            The nodes dictionary of the circuit, so that the method
            can convert its internal node IDs to the corresponding
            external ones.

        **Returns:**

        ntlst_line : string
            The netlist line.
        """
        return "%s %s %s %s %s %g" % (self.part_id, nodes_dict[self.n1],
                                nodes_dict[self.n2], nodes_dict[self.sn1],
                                nodes_dict[self.sn2], self.alpha)


class GISource(Component):

    """Linear voltage controlled current source

    .. image:: images/elem/vccs.svg

    The source port is an open circuit, the output port is an ideal current
    source:

    .. math::

        \\left\\{
        \\begin{array}{ll}
            I_s = 0\\\\
            I_o = \\alpha \\cdot (V(sn_1) - V(sn_2))
        \\end{array}
        \\right.


    Where :math:`I_s` is the current at the source port and :math:`I_o` is the
    current at the output port.
    The remaining symbols are explained in the Parameters section below.

    .. note::

        This simulator uses the passive convention: a positive current flows
        into the element through the anode and exits through the cathode.

    **Parameters:**

    n1 : int
        *Internal* node to be connected to the anode of the output port.
    n2 : int
        *Internal* node to be connected to the cathode of the output port.
    value : float
        Proportionality constant :math:`\\alpha` between the sense voltage and
        the output current, in Ampere/Volt.
    sn1 : int
        *Internal* node to be connected to the anode of the source (sensing)
        port.
    sn2 : int
        *Internal* node to be connected to the cathode of the source
        (sensing) port.

    """

    def __init__(self, part_id, n1, n2, value, sn1, sn2):
        self.part_id = part_id
        self.n1 = n1
        self.n2 = n2
        self.alpha = value
        self.sn1 = sn1
        self.sn2 = sn2
        self.is_nonlinear = False
        self.is_symbolic = True

    def __str__(self):
        return "value=%s" % self.alpha

    def get_netlist_elem_line(self, nodes_dict):
        """A netlist line that, parsed, evaluates to the same instance

        **Parameters:**

        nodes_dict : dict
            The nodes dictionary of the circuit, so that the method
            can convert its internal node IDs to the corresponding
            external ones.

        **Returns:**

        ntlst_line : string
            The netlist line.
        """
        return "%s %s %s %s %s %g" % (self.part_id, nodes_dict[self.n1],
                                nodes_dict[self.n2], nodes_dict[self.sn1],
                                nodes_dict[self.sn2], self.alpha)


class HVSource(Component):  # H

    """Linear current-controlled voltage source

    .. image:: images/elem/ccvs.svg

    The source port is an existing voltage source, used to sense the current
    controlling the voltage source connected to the destination port.

    Mathematically, it is equivalent to the following:

    .. math::

        \\left\\{
        \\begin{array}{ll}
            V(sn_1) - V(sn_2) = V_S \\\\
            Vn_1 - Vn_2 = \\alpha * I[V_s]
        \\end{array}
        \\right.

    Where :math:`I[V_s]` is the current flowing in the source port, :math:`V_s`
    is the voltage applied between the nodes :math:`sn_1` and :math:`sn_2`.
    The remaining symbols are explained in the Parameters section below.

    .. note::

        This simulator uses the passive convention: a positive current flows
        into the element through the anode and exits through the cathode.

    **Parameters:**

    n1 : int
        *Internal* node to be connected to the anode of the output port.
    n2 : int
        *Internal* node to be connected to the cathode of the output port.
    value : float
        Proportionality constant :math:`\\alpha` between the sense current and
        the output voltage, in V/A.
    source_id : string
        ``part_id`` of the current-sensing voltage source, eg. ``'V1'`` or
        ``'VSENSE'``.

    """

    def __init__(self, part_id, n1, n2, value, source_id):
        self.part_id = part_id
        self.n1 = n1
        self.n2 = n2
        self.alpha = value
        self.source_id = source_id
        self.is_nonlinear = False
        self.is_symbolic = True

    def get_netlist_elem_line(self, nodes_dict):
        """A netlist line that, parsed, evaluates to the same instance

        **Parameters:**

        nodes_dict : dict
            The nodes dictionary of the circuit, so that the method
            can convert its internal node IDs to the corresponding
            external ones.

        **Returns:**

        ntlst_line : string
            The netlist line.
        """
        return "%s %s %s %s %g" % (self.part_id, nodes_dict[self.n1],
                                nodes_dict[self.n2], self.source_id,
                                self.alpha)


class FISource(Component):
    """Linear current-controlled current source

    .. image:: images/elem/cccs.svg

    This element implements a current source whose current value is controlled
    by the current flowing in a current source, which acts as the "sensing"
    element.

    Mathematically:

    .. math::


        \\left\\{
        \\begin{array}{ll}
            V(sn_1) - V(sn_2) = V_S \\\\
            I_o = \\alpha \\cdot I_s
        \\end{array}
        \\right.


    Where :math:`V_s` is the voltage forced at the source port by the sensing
    element and :math:`I_o` is the current at the output port.
    The remaining symbols are explained in the Parameters section below.

    .. note::

        This simulator uses the passive convention: a positive current flows
        into the element through the anode and exits through the cathode.

    **Parameters:**

    n1 : int
        *Internal* node to be connected to the anode of the output port.
    n2 : int
        *Internal* node to be connected to the cathode of the output port.
    value : float
        Proportionality constant :math:`\\alpha` between the sense current and
        the output current.
    source_id : string
        ``part_id`` of the sensing voltage source, eg. ``'V1'`` or ``'VSENSE'``.

    """
    #F
    def __init__(self, part_id, n1, n2, value, source_id):
        self.part_id = part_id
        self.n1 = n1
        self.n2 = n2
        self.source_id = source_id
        self.alpha = value
        self.is_nonlinear = False
        self.is_symbolic = True

    def get_netlist_elem_line(self, nodes_dict):
        """A netlist line that, parsed, evaluates to the same instance

        **Parameters:**

        nodes_dict : dict
            The nodes dictionary of the circuit, so that the method
            can convert its internal node IDs to the corresponding
            external ones.

        **Returns:**

        ntlst_line : string
            The netlist line.
        """
        return "%s %s %s %s %g" % (self.part_id, nodes_dict[self.n1],
                                nodes_dict[self.n2], self.source_id,
                                self.alpha)




# -*- coding: iso-8859-1 -*-
# pss.py
# Periodic Steady State analysis
# Copyright 2011-2013 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

from . import shooting
from . import bfpss
from . import options

# Periodic Steady State (PSS) analysis module.

# This module is an interface to a generic PSS analysis,
# which will be set up automatically for you according to
# the algorithm selected.

# Netlist syntax
# .PSS PERIOD=n [points=n step=n autonomous=bool method=str]

specs = {'pss': {'tokens': ({
                            'label': 'period',
                            'pos': 0,
                            'type': float,
                            'needed': True,
                            'dest': 'period',
                            'default': None
                            },
        {
                            'label': 'points',
                            'pos': None,
                            'type': float,
                            'needed': True,
                            'dest': 'points',
                            'default': None
                            },
                 {
                            'label': 'step',
                            'pos': None,
                            'type': float,
                            'needed': False,
                            'dest': 'step',
                            'default': None
                            },
                 {
                            'label': 'autonomous',
                            'pos': None,
                            'type': bool,
                            'needed': False,
                            'dest': 'autonomous',
                            'default': '0'
                            },
                 {
                            'label': 'method',
                            'pos': None,
                            'type': str,
                            'needed': False,
                            'dest': 'method',
                            'default': options.SHOOTINGPSS
                            }
                            )
                 }
         }


def pss_analysis(*largs, **args):
    """Perform a PSS analysis.

    The only required argument is ``method`` (string), which
    selects the algorithm to be used.

    Two algorithms are a shooting PSS, selected with the ``"shooting"`` switch
    and brute-force PSS, selected by the ``"brute-force"`` switch. Any other
    value for the ``method`` parameter will result in a ``ValueError`` exception
    being raised.

    The rest of the arguments will be passed to the algorithm implementing the
    analysis.

    For the shooting algorithm see :mod:`ahkab.shooting`, for the brute-force
    algorithm see :mod:`ahkab.bfpss`.

    **Returns:**
    
    sol : PSS solution object (:class:`results.pss_solution`)
        The solution.
    """
    m = args.pop('method').lower()
    if m == options.SHOOTINGPSS:
        r = shooting.shooting_analysis(*largs, **args)
    elif m == options.BFPSS:
        r = bfpss.bfpss(*largs, **args)
    else:
        raise Exception("Unknown PSS method %s" % m)
    return r



# -*- coding: iso-8859-1 -*-
# implicit_euler.py
# Implicit euler DF
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""This module implements the Implicit Euler (IE, aka Backward Euler, BE) and a
first-order forward formula (FF) to be used for prediction.

The formula is:

.. math::

    x'_{n+1} = C_0 x_{n+1} + C_1 x_{n}

Where:

* :math:`C_0 = 1/h`
* :math:`C_1 = -1/h`

The backward Euler method is not only A-stable, making it suitable for the solution of
stiff equations but is even L-stable.

Module reference
================

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import numpy as np

#: The order of the differentiation formula
order = 1

def is_implicit():
    """Is this differentiation formula implicit?"""
    return True


def get_df_coeff(step):
    """Get the coefficients for a Backward Euler differentiation step

    The first coefficient is the factor for the new point :math:`x_{n+1}`,
    the second is the one for the previous point :math:`x_{n}`.

    If the step value is :math:`h`, this method returns:

    .. math::

        [1/h, -1/h]

    **Parameters:**

    step : float
        The differentiation formula step value.

    **Returns:**

    c0, c1 : floats
        The coefficients of :math:`x_{n+1}` and :math:`x_{n}`.
    """
    return 1.0/step, -1.0/step


def get_required_values():
    """Get what values are required by the DF and the FF

    **Returns**

    The method returns two tuples, each of them having the form:

    ``[max_order_of_x, max_order_of_dx]``

    The first tuple is the one to be considered if no Forward Formula (FF) is
    needed, the second if the FF is also required.

    Both the values in each tuple can be either of type ``int`` or be set to
    ``None``.

    If ``max_order_of_x`` is set to an arbitrary positive integer value
    :math:`k`, the Differentiation Formula (DF) needs all the :math:`x_{n-i}`
    values of :math:`x`, where :math:`i \le k` (the value x has :math:`i+1`
    steps before the one we will ask for the derivative).  The same applies to
    ``max_order_of_dx``, but it regards :math:`dx/dt` instead of :math:`x`.

    If ``max_order_of_x`` or ``max_order_of_dx`` are set to ``None``, that means
    that no value of :math:`x`, or :math:`dx/dt`, is required.

    In the case at hand, where the formula is the Backward Euler (BE, aka
    Implicit Euler, IE), this method will return:

    ``((0, None), (1, None))``

    """
    # we need just the x(n) value
    return ((0, None), (1, None))


def has_ff():
    """Is a forward formula for prediction available?

    **Returns:**

    ``True``

    """
    return True


def get_df(pv_array, suggested_step, predict=True):
    """Get the coefficients for the DF, FF and LTE calculation

    **Parameters:**

    pv_array : list
        It must be an list of lists, each of them having the structure
        ``[time, xnk, dxnk]``.

        In particular, the ``pv_array[k]`` element of ``pv_array`` is composed
        of:

        * ``time``, float, which is the time at which the solution is valid:
          ``t(n-k)``,
        * ``xnk``, ndarray, which is :math:`x_{n-k}`,
        * ``dxnk``, ndarray, :math:`dx_{n-k}/dt`.

        The length of ``pv_array`` has to match the value returned by
        :func:`get_required_values`.

        Any values that are not needed may be set to ``None``, and they will be
        disregarded.

    suggested_step : float
        The step that is (expected) to be used in the DF. It is only an
        expectation because it may be rejected at a later stage if there is step
        control enabled.
    predict : boolean, optional
        Whether the terms for a prediction formula are required as well or not.
        Defaults to ``True``.

    **Returns:**

    ret : tuple
        ``ret`` is a tuple of 5 elements, where:

        * the [0] element is the coeffiecient of :math:`x_{n+1}` (scalar),
        * the [1] element is the matrix of constant terms of shape ``(Nx1)`` of
          :math:`x_{n+1}`,
        * the [2] element is the coefficient of the LTE of :math:`x_{n+1}` (scalar),
        * the [3] element is the predicted value of :math:`x_{n+1}` (matrix), only
          available if the ``predict`` parameter is set to ``True``. Otherwise it's
          ``None``.
        * the [4] element is the coefficient of the LTE of the prediction (matrix),
          also only available if the ``predict`` parameter is set to ``True``,
          otherwise, it is ``None``.

    .. note::

        With the returned values, the derivative may then be written as:

        .. math::

            \\frac{dx_{n+1}}{dt} = \\mathrm{ret[0]}\; x_{n+1} + \\mathrm{ret[1]}

    """

    # we'll use just the first column, since our method needs just x(n)
    if len(pv_array[0]) != 3:
        raise ValueError('Malformed pv_array: length == %d (3)' % len(pv_array))
    if pv_array[0][1] is None:
        raise ValueError('Malformed pv_array: pv_array[0][1] is None')

    x_lte_coeff = 0.5*suggested_step

    if predict and len(pv_array) > 1 and pv_array[1][1] is not None:
        predict = np.zeros(pv_array[0][1].shape)
        for index in range(predict.shape[0]):
            predict[index, 0] = (pv_array[0][1][index, 0] -
                                 pv_array[1][1][index, 0]) \
                                / (pv_array[0][0] -
                                   pv_array[1][0])*suggested_step \
                                + pv_array[0][1][index, 0]
        predict_lte_coeff = -0.5*suggested_step*\
                            (pv_array[0][0] + suggested_step - pv_array[1][0])
    else:
        predict = None
        predict_lte_coeff = None

    return (1./suggested_step, -1./suggested_step*pv_array[0][1],
            x_lte_coeff, predict, predict_lte_coeff)



#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
# __main__.py
# The frontend of the simulator
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, division

import sys
from optparse import OptionParser

from . import ahkab
from . import options
from . import transient
from . import utilities
from .__version__ import __version__

def _cli():
    usage = "usage: \t%prog [options] <filename>\n\nThe filename is the " + \
            "netlist to be open. Use - (a dash) to read from stdin."
    parser = OptionParser(usage, version="%prog " + __version__ + \
             " (c) 2006-2015 Giuseppe Venturini")

    # general options
    parser.add_option("-v", "--verbose", action="store", type="string",
                     dest="verbose", default="3", help="Verbose level: " +
                    "from 0 (almost silent) to 5 (debug)")
    parser.add_option("-p", "--print", action="store_true",
                      dest="print_circuit", default=False, help="Print " +
                      "the parsed circuit")
    parser.add_option("-o", "--outfile", action="store", type="string",
                      dest="outfile", default='stdout', help="Data output " +
                      "file. Defaults to stdout.")
    parser.add_option("", "--dc-guess", action="store", type="string",
                      dest="dc_guess", default="guess", help="Guess to be " +
                      "used to start a OP or DC analysis: none or guess. " +
                      "Defaults to guess.")
    parser.add_option("-t", "--tran-method", action="store", type="string",
                      dest="method", default=transient.TRAP.lower(),
                      help="Method to be used in transient analysis: " +
                      transient.IMPLICIT_EULER.lower() + ", " +
                      transient.TRAP.lower() + ", " +
                      transient.GEAR2.lower() + ", " +
                      transient.GEAR3.lower() + ", " +
                      transient.GEAR4.lower() + ", " +
                      transient.GEAR5.lower() + " or " +
                      transient.GEAR6.lower() + ". Defaults to TRAP.")
    parser.add_option("", "--t-fixed-step", action="store_true",
                      dest="no_step_control", default=False, help="Disables" +
                      " the step control in transient analysis.")
    parser.add_option("", "--v-absolute-tolerance", action="store",
                      type="string", dest="vea", default=None, help="Voltage" +
                      " absolute tolerance. Default: " + str(options.vea) + " V")
    parser.add_option("", "--v-relative-tolerance", action="store",
                      type="string", dest="ver", default=None, help="Voltage " +
                      "relative tolerance. Default: " + str(options.ver))
    parser.add_option("", "--i-absolute-tolerance", action="store",
                      type="string", dest="iea", default=None, help="Current " +
                      "absolute tolerance. Default: " + str(options.iea) + " A")
    parser.add_option("", "--i-relative-tolerance", action="store",
                      type="string", dest="ier", default=None, help="Current " +
                      "relative tolerance. Default: " + str(options.ier))
    parser.add_option("", "--h-min", action="store", type="string",
                      dest="hmin", default=None, help="Minimum time step. " +
                      "Default: " + str(options.hmin))
    parser.add_option("", "--dc-max-nr", action="store", type="string",
                      dest="dc_max_nr_iter", default=None, help="Maximum " +
                      "number of NR iterations for DC and OP analyses. " +
                      "Default: " + str(options.dc_max_nr_iter))
    parser.add_option("", "--t-max-nr", action="store", type="string",
                      dest="transient_max_nr_iter", default=None,
                      help="Maximum number of NR iterations for each time " +
                      "step during transient analysis. Default: " +
                      str(options.transient_max_nr_iter))
    parser.add_option("", "--t-max-time", action="store", type="string",
                      dest="transient_max_time_iter", default=None,
                      help="Maximum number of time iterations during " +
                      "transient analysis. Setting it to 0 (zero) disables " +
                      "the limit. Default: " +
                      str(options.transient_max_time_iter))
    parser.add_option("", "--s-max-nr", action="store", type="string",
                      dest="shooting_max_nr_iter", default=None,
                      help="Maximum number of NR iterations during shooting " +
                      "analysis. Setting it to 0 (zero) disables the " +
                      "limit. Default: " + str(options.shooting_max_nr_iter))
    parser.add_option("", "--gmin", action="store", type="string", dest="gmin",
                      default=None, help="The minimum conductance to ground. " +
                      "Inserted when requested. Default: " +
                      str(options.gmin))
    parser.add_option("", "--cmin", action="store", type="string", dest="cmin",
                      default=None, help="The minimum capacitance to ground. " +
                      "Default: " + str(options.cmin))

    (cli_options, remaning_args) = parser.parse_args()

    verbose = int(cli_options.verbose)
    if cli_options.method is not None:
        options.default_tran_method = cli_options.method.upper()
    if cli_options.vea is not None:
        options.vea = float(cli_options.vea)
    if cli_options.ver is not None:
        options.ver = float(cli_options.ver)
    if cli_options.iea is not None:
        options.iea = float(cli_options.iea)
    if cli_options.ier is not None:
        options.ier = float(cli_options.ier)
    if cli_options.hmin is not None:
        options.hmin = float(cli_options.hmin)
    if cli_options.dc_max_nr_iter is not None:
        options.dc_max_nr_iter = int(float(cli_options.dc_max_nr_iter))
    if cli_options.transient_max_nr_iter is not None:
        options.transient_max_nr_iter = int(
            float(cli_options.transient_max_nr_iter))
    if cli_options.transient_max_time_iter is not None:
        options.transient_max_time_iter = int(
            float(cli_options.transient_max_time_iter))
    if cli_options.shooting_max_nr_iter is not None:
        options.shooting_max_nr_iter = int(
            float(cli_options.shooting_max_nr_iter))
    if cli_options.gmin is not None:
        options.gmin = float(cli_options.gmin)
    if cli_options.cmin is not None:
        options.cmin = float(cli_options.cmin)

    if not len(remaning_args) == 1:
        print("Usage: ./ahkab.py [options] <filename>\n./ahkab.py -h for help")
        sys.exit(1)
    if remaning_args[0] == '-':
        read_netlist_from_stdin = True
    else:
        read_netlist_from_stdin = False
    if not read_netlist_from_stdin and not utilities.check_file(remaning_args[0]):
        sys.exit(23)

    options.transient_no_step_control = cli_options.no_step_control
    options.dc_use_guess = cli_options.dc_guess
    options.cli = True
    ahkab._print = cli_options.print_circuit

    # Program execution
    ahkab.main(filename=remaning_args[0],
               outfile=cli_options.outfile, verbose=verbose)

    sys.exit(0)

if __name__ == "__main__":
    _cli()



# -*- coding: iso-8859-1 -*-
# constants.py
# To hold the various constants required
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""Constants useful for building equations and expressions describing
semiconductor physics
"""


from __future__ import (unicode_literals, absolute_import,
                        division, print_function)
import math

#: The electron charge :math:`e`.
e = 1.60217646e-19
#: Simulation temperature in Kelvin degrees.
T = 300
#: Reference temperature in Kelvin degrees.
Tref = 300
#: The Boltzmann constant
k = 1.3806503e-23


def Vth(T=Tref):
    """The thermal voltage: :math:`kT/q`.

    **Parameters:**

    T : float, optional
        The temperature at which the thermal voltage is to be evaluated.
        If not specified, it defaults to ``constants.Tref``.

    **Returns:**

    vth : float
        The thermal voltage, :math:`kT/q`.
    """
    return k * T / e


class silicon:
    """Silicon class

    Access this class as ``constants.si``.

    **Attributes**

    **esi**: permittivity of silicon.

    **eox**: permittivity of silicon dioxide.

    """
    def __init__(self):
        self.esi = 104.5 * 10 ** -12  # F/m
        self.eox = 34.5 * 10 ** -12  # F/m

    def Eg(self, T=Tref):
        """Energy gap of silicon at temperature ``T``

        **Parameters:**

        T : float, optional
            The temperature at which the thermal voltage is to be evaluated.
            If not specified, it defaults to ``constants.Tref``.

        **Returns:**

        Eg : float
            The energy gap, expressed in electron-volt (eV).
        """

        return (1.16 - 0.000702 * T ** 2 / (1108 + T))  # eV

    def ni(self, T=Tref):
        """Intrinsic Silicon carrier concentration at temperature ``T``

        **Parameters:**

        T : float, optional
            The temperature at which the thermal voltage is to be evaluated.
            If not specified, it defaults to ``constants.Tref``.

        **Returns:**

        ni : float
            The intrinsec carrier concentration.
        """

        return 1.45 * 10 ** 16 * (T / Tref) * math.exp(self.Eg(Tref) / (2 * Vth(Tref)) - self.Eg(T) / (2 * Vth(T)))

#: Silicon class instantiated.
si = silicon()



# -*- coding: iso-8859-1 -*-
# ticker.py
# Progress indicator class
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
A progress indicator.
"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import sys

class ticker:

    """This is a progress indicator class.

    If activated, you shouldn't print anything to screen before calling
    ticker.hide().

    If you wish to change the progress indicator, change self.progress to
    something else.
    """
    def __init__(self, increments_for_step=10):
        self.progress = ("-", "\\", "|", "/")
        self._index = 0
        self._step = 0
        self._display = False
        self.increments_for_step = increments_for_step

    def step(self):
        """After calling this function ticker.increments_for_step times
        the status is incremented."""
        if not self._display:
            return
        if (self._index + 1) % self.increments_for_step == 0:
            if (self._step + 1) % len(self.progress) == 0:
                self._step = 0
            else:
                self._step = self._step + 1
            self._index = 0
            if self._display:
                sys.stdout.write("\b" + self.progress[self._step])
                sys.stdout.flush()
        else:
            self._index = self._index + 1

    def hide(self, enable=None):
        """Before printing text to screen, call this to hide the progress
        indicator.
        """
        if enable == False:
            return
        sys.stdout.write("\b")
        sys.stdout.flush()
        self._display = False

    def display(self, enable=None):
        """Print to screen the progress indicator. Call hide to hide it
        again.
        """
        if enable == False:
            return
        sys.stdout.write(self.progress[self._step])
        sys.stdout.flush()
        self._display = True

    def reset(self):
        """Reset to initial status. Doesn't hide it."""
        self._step = 0
        self._index = 0

if __name__ == "__main__":
    # test
    import time
    tk = ticker()
    tk.display(True)
    for i in range(1000):
        time.sleep(0.005)
        tk.step()
    tk.hide()



#!/usr/bin/env python
import os
import re
from setuptools import setup, find_packages

with open('ahkab/__version__.py') as version_file:
    __version__ = re.findall("__version__ = '([^']*)'", version_file.read())[0]

def read(fname):
    try:
        with open(os.path.join(os.path.dirname(__file__), fname)) as fp:
            return fp.read()
    except IOError:
        return ""

setup(
    name='ahkab',
    version=__version__,
    packages=find_packages(),
    package_data={'ahkab': ['test_data/*.dat']},
    install_requires=['numpy>1.7.0', 'scipy>=0.14.0', 'sympy>=0.7.6',
                      'tabulate>=0.7.3'],
    extras_require = {'PLOT':['matplotlib>=1.1.1']},
    zip_safe=False,
    include_package_data=True,
    author="Giuseppe Venturini and others",
    author_email="ggventurini+GITHUB@gmail.com",
    description="a SPICE-like electronic circuit simulator",
    long_description=''.join([read('pypi_description.rst'), '\n\n',
                              read('CHANGES.rst')]),
    entry_points = {'console_scripts': ['ahkab = ahkab.__main__:_cli', ], },
    license="GPL",
    keywords="electronic circuit simulator numeric symbolic",
    url='http://ahkab.github.io/ahkab/',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4"]
)




# -*- coding: iso-8859-1 -*-
# ekv.py
# Partial implementation of the EKV 3.0 MOS transistor model
# Copyright 2010 Giuseppe Venturini
#
# The EKV model was developed by Matthias Bucher, Christophe Lallement,
# Christian Enz, Fabien Théodoloz, François Krummenacher at the Electronics
# Laboratories, Swiss Federal Institute of Technology (EPFL),
# Lausanne, Switzerland.
# This implementation is based upon:
# 1. Matthias Bucher, Christian Enz, François Krummenacher, Jean-M. Sallese,
# Christophe Lallement and Alain-S. Porret,
# The EKV 3.0 Compact MOS Transistor Model: Accounting for Deep-Submicron
# Aspects, <http://www.nsti.org/publications/MSM/2002/pdf/346.pdf>
# 2. EKV 2.6 Technical report, <http://legwww.epfl.ch/ekv/pdf/ekv_v262.pdf>.
#
# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
Partial implementation of the EKV 3.0 MOS transistor model

The EKV model was developed by Matthias Bucher, Christophe Lallement,
Christian Enz, Fabien Théodoloz, François Krummenacher at the Electronics
Laboratories, Swiss Federal Institute of Technology (EPFL),
Lausanne, Switzerland.

The Tecnical Report upon which this implementation
is based is available here:

`EKV Technical Report <http://legwww.epfl.ch/ekv/pdf/ekv_v262.pdf>`_.

This module defines two classes:

* :class:`ekv_device`
* :class:`ekv_mos_model`


**Features:**

* EKV model implementation, computation of charges, potentials,
  reverse and forward currents, slope factor and normalization factors.
* Calculation of trans-conductances based on the charge-driven approach.
* N/P MOS symmetry
* Rudimentary temperature effects.

**The Missing Features:**

* Channel length modulation,
* Reverse Short Channel Effect (RSCE),
* Complex mobility degradation,
* Transcapacitances,
* Quasi-static implementation,

Patches to implement the above are welcome!

.. note::
    The default values in the model are suitable for a generic
    500nm feature-size CMOS process.

"""

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import scipy, scipy.optimize
import math

from scipy.optimize import newton

from . import constants
from . import options
from . import utilities
from . import printing


# DEFAULT VALUES FOR 500n CH LENGTH
COX_DEFAULT = .7e-3
VTO_DEFAULT = .5
GAMMA_DEFAULT = 1.0
PHI_DEFAULT = .7
KP_DEFAULT = 50e-6
UCRIT_DEFAULT = 2e6
LAMBDA_DEFAULT = .5
XJ_DEFAULT = .1e-6


TCV_DEFAULT = 1e-3
BEX_DEFAULT = -1.5

ISMALL_GUESS_MIN = 1e-10


class ekv_device:
    """EKV device

    **Parameters:**

    part_id : string
        The element identifier, eg 'M1'

    nd : int
        drain node
    ng : int
        gate node
    ns : int
        source node
    nb : int
        bulk node
    L : float
        element width [m]
    W : float
        element length [m]
    M : int
        multiplier (n. of shunt devices)
    N : int
        series mult. (n. of series devices)
    model : ekv_model instance
        The corresponding instance of ekv_mos_model

    Selected methods:
    - get_output_ports() -> (nd, ns)
    - get_drive_ports() -> (nd, nb), (ng, nb), (ns, nb)

    """

    INIT_IFRN_GUESS = 1

    def __init__(self, part_id, nd, ng, ns, nb, W, L, model, M=1, N=1):
        self.ng = ng
        self.nb = nb
        self.n1 = nd
        self.n2 = ns
        self.ports = ((self.n1, self.nb), (
            self.ng, self.nb), (self.n2, self.nb))

        class dev_class:
            pass  # empty class to hold device parameters
        self.device = dev_class()
        self.device.L = float(L)  # channel length -
        self.device.W = float(W)  # channel width -
        self.device.M = int(M)  # parallel multiple device number
        self.device.N = int(N)  # series multiple device number
        self.ekv_model = model
        self.opdict = {}
        self.opdict.update(
            {'state': (float('nan'), float('nan'), float('nan'))})
        self.opdict.update({'ifn': self.INIT_IFRN_GUESS})
        self.opdict.update({'irn': self.INIT_IFRN_GUESS})
        self.opdict.update(
            {'ip_abs_err': self.ekv_model.get_ip_abs_err(self.device)})
        self.part_id = part_id
        self.is_nonlinear = True
        self.is_symbolic = True
        self.dc_guess = [self.ekv_model.VTO * (
            0.1) * self.ekv_model.NPMOS, self.ekv_model.VTO * (1.1) * self.ekv_model.NPMOS, 0]

        devcheck, reason = self.ekv_model._device_check(self.device)
        if not devcheck:
            raise Exception(reason + " out of boundaries.")

    def get_drive_ports(self, op):
        """Returns a tuple of tuples of ports nodes, as:
        (port0, port1, port2...)
        Where each port is in the form:
        port0 = (nplus, nminus)
        """
        return self.ports  # d,g,s

    def get_output_ports(self):
        return ((self.n1, self.n2),)

    def __str__(self):
        mos_type = self._get_mos_type()
        rep = " " + self.ekv_model.name + " w=" + str(self.device.W) + " l=" + \
            str(self.device.L) + " M=" + str(self.device.M) + " N=" + \
            str(self.device.N)

        return rep

    def _get_mos_type(self):
        """Returns N or P (capitalized)
        """
        mtype = 'N' if self.ekv_model.NPMOS == 1 else 'P'
        return mtype

    def i(self, op_index, ports_v, time=0):
        """Returns the current flowing in the element with the voltages
        applied as specified in the ports_v vector.

        ports_v: [voltage_across_port0, voltage_across_port1, ...]
        time: the simulation time at which the evaluation is performed.
              It has no effect here. Set it to None during DC analysis.

        """
        ret, j1, j2 = self.ekv_model.get_ids(self.device, ports_v,
                                             self.opdict)

        return ret

    def update_status_dictionary(self, ports_v):
        if self.opdict is None:
            self.opdict = {}
        if not (self.opdict['state'] == ports_v[0] and 'gmd' in self.opdict) or \
            not (self.opdict['state'] == ports_v[0] and 'gmg' in self.opdict) or \
            not (self.opdict['state'] == ports_v[0] and 'gms' in self.opdict) or \
                not (self.opdict['state'] == ports_v[0] and 'Ids' in self.opdict):

            self.opdict['state'] == ports_v[0]
            self.opdict['gmd'] = self.g(0, ports_v[0], 0)
            self.opdict['gmg'] = self.g(0, ports_v[0], 1)
            self.opdict['gms'] = self.g(0, ports_v[0], 2)
            self.opdict['Ids'] = self.i(0, ports_v[0])

        gmd = self.opdict['gmd']
        gmg = self.opdict['gmg']
        gms = self.opdict['gms']
        ids = self.opdict['Ids']

        if ids == 0:
            TEF = float('nan')
        else:
            TEF = abs(gms * constants.Vth() / ids)
        self.opdict['TEF'] = TEF

    def get_op_info(self, ports_v):
        """Information regarding the Operating Point (OP)

        **Parameters:**

        ports_v : list of lists
            The voltages applied to all the driving ports, grouped by output
            port.

        i.e.

        ::

            [<list of voltages for the drive ports of output port 0>,
             <list of voltages for the drive ports of output port 1>,
             ...,
             <list of voltages for the drive ports of output port N>]

        Usually, this method returns ``op_keys`` and the corresponding
        ``op_info``, two lists, one holding the labels, the other the
        corresponding values.

        In the case of MOSFETs, the values are way too many to be shown in a
        linear table. For this reason, we return ``None`` as ``op_keys``, and we
        return for ``op_info`` a list which holds both labels and values in a
        table-like manner, spanning the vertical and horizontal dimension.

        For this reason, each MOSFET has to have its OP info printed alone, not
        grouped as it happens with most other elements.

        **Returns:**

        op_keys : ``None``
            See above for why this value is always ``None``.
        op_info : list of floats
            The OP information ready to be passed to :func:`printing.table` for
            arranging it in a pretty table to display.
        """
        mos_type = self._get_mos_type()

        self.update_status_dictionary(ports_v)

        sat_status = "SATURATION" if self.opdict['SAT'] else "LINEAR"
        if self.opdict["WMSI"] == 0:
            wmsi_status = "WEAK INVERSION"
        if self.opdict["WMSI"] == 1:
            wmsi_status = "MODERATE INVERSION"
        if self.opdict["WMSI"] == 2:
            wmsi_status = "STRONG INVERSION"

        arr = [
            [self.part_id, mos_type.upper() + " ch", wmsi_status, "", "", sat_status, "", "", "", "", "", ""], ]
        arr.append(
            ["beta", "[A/V^2]:", self.opdict['beta'], "Weff", "[m]:", str(self.opdict['Weff']) + " (" + str(self.device.W) + ")",
             "Leff", "[m]:", str(self.opdict['Leff']) + " (" + str(self.device.L) + ")", "M/N:", "", str(self.device.M) + "/" + str(self.device.N)])
        arr.append(
            ["Vdb", "[V]:", float(ports_v[0][0]), "Vgb", "[V]:", float(ports_v[0][1]),
             "Vsb", "[V]:", float(ports_v[0][2]),  "Vp", "[V]:", self.opdict['Vp'], ])
        arr.append(
            ["VTH", "[V]:", self.opdict['VTH'], "VOD", "[V]:", self.opdict['VOD'],
             "nq: ", "", self.opdict['nq'], "VA", "[V]:", str(self.opdict['Ids'] / self.opdict['gmd'])])
        arr.append(
            ["Ids", "[A]:", self.opdict['Ids'], "nv: ", "", self.opdict['nv'],
             "Ispec", "[A]:", self.opdict["Ispec"], "TEF:", "", str(self.opdict['TEF']), ])
        arr.append(["gmg", "[S]:", self.opdict['gmg'], "gms", "[S]:",
                   self.opdict['gms'], "rob", u"[\u2126]:", 1 / self.opdict['gmd'], "", "", ""])
        arr.append(
            ["if:", "", self.opdict['ifn'], "ir:", "", self.opdict['irn'],
             "Qf", "[C/m^2]:", self.opdict["qf"], "Qr", "[C/m^2]:", self.opdict["qr"], ])
        # arr.append([  "", "", "", "", "", ""])

        return None, arr

    def g(self, op_index, ports_v, port_index, time=0):
        """Returns the differential (trans)conductance rs the port specified by port_index
        when the element has the voltages specified in ports_v across its ports,
        at (simulation) time.

        ports_v: a list in the form: [voltage_across_port0, voltage_across_port1, ...]
        port_index: an integer, 0 <= port_index < len(self.get_ports())
        time: the simulation time at which the evaluation is performed. Set it to
        None during DC analysis.
        """

        assert op_index == 0
        assert port_index < 3

        if port_index == 0:
            g = self.ekv_model.get_gmd(self.device, ports_v, self.opdict)
        elif port_index == 1:
            g = self.ekv_model.get_gmg(self.device, ports_v, self.opdict)
        if port_index == 2:
            g = self.ekv_model.get_gms(self.device, ports_v, self.opdict)

        if op_index == 0 and g == 0:
            if port_index == 2:
                sign = -1
            else:
                sign = +1
            g = sign * options.gmin * 2.0

        # print type(g), g

        if op_index == 0 and port_index == 0:
            self.opdict.update({'gmd': g})
        elif op_index == 0 and port_index == 1:
            self.opdict.update({'gmg': g})
        elif op_index == 0 and port_index == 2:
            self.opdict.update({'gms': g})

        return g

    def get_value_function(self, identifier):
        def get_value(self):
            return self.opdict[identifier]
        return get_value

    def get_netlist_elem_line(self, nodes_dict):
        mos_type = self._get_mos_type()
        return "%s %s %s %s %s %s type=%s w=%g l=%g m=%g n=%g" % \
              (self.part_id, nodes_dict[self.n1], nodes_dict[self.ng],
              nodes_dict[self.n2], nodes_dict[self.nb], self.ekv_model.name,
              mos_type, self.device.W, self.device.L, self.device.M,
              self.device.N)


class scaling_holder:
    pass  # will hold the scaling factors


class ekv_mos_model:

    def __init__(self, name=None, TYPE='n', TNOM=None, COX=None,
                 GAMMA=None, NSUB=None, PHI=None, VTO=None, KP=None,
                 XJ=None, LAMBDA=None,
                 TOX=None, VFB=None, U0=None, TCV=None, BEX=None):

        self.scaling = scaling_holder()

        self.name = "model_ekv0" if name is None else name
        self.TNOM = float(TNOM) if TNOM is not None else constants.Tref
        self.NPMOS = 1 if TYPE == 'n' else -1

        # optional parameters (no defaults)
        self.TOX = float(TOX) if TOX is not None else None
        self.NSUB = float(NSUB) if NSUB is not None else None
        self.VFB = self.NPMOS * float(VFB) if VFB is not None else None
        self.U0 = float(U0) if U0 is not None else None

        # crucial parameters
        if COX is not None:
            self.COX = float(COX)
        elif TOX is not None:
            self.COX = constants.si.eox / TOX
        else:
            self.COX = COX_DEFAULT

        if GAMMA is not None:
            self.GAMMA = float(GAMMA)
        elif NSUB is not None:
            self.GAMMA = math.sqrt(2.0*constants.e*constants.si.esi*NSUB
                                   *10**6/self.COX)
        else:
            self.GAMMA = GAMMA_DEFAULT
        if PHI is not None:
            self.PHI = float(PHI)
        elif NSUB is not None:
            self.PHI = 2. * constants.Vth(self.TNOM) * \
                       math.log(NSUB*10.0**6.0/constants.si.ni(self.TNOM))
        else:
            self.PHI = PHI_DEFAULT
        if VTO is not None:
            self.VTO = self.NPMOS * float(VTO)
            if self.VTO < 0:
                printing.print_warning("model %s has internal negative VTO (%f V)."
                                       % (self.name, self.VTO))
        elif VFB is not None:
            self.VTO = VFB + PHI + GAMMA * PHI  # inv here??
        else:
            self.VTO = self.NPMOS * VTO_DEFAULT

        if KP is not None:
            self.KP = float(KP)
        elif U0 is not None:
            self.KP = (U0 * 10.0 ** -4) * self.COX
        else:
            self.KP = KP_DEFAULT

        self.LAMBDA = LAMBDA if LAMBDA is not None else LAMBDA_DEFAULT
        self.XJ = XJ if XJ is not None else XJ_DEFAULT
        self.UCRIT = UCRIT_DEFAULT
        # Intrinsic model temperature parameters
        self.TCV = self.NPMOS * \
            float(TCV) if TCV is not None else self.NPMOS * TCV_DEFAULT
        self.BEX = float(BEX) if BEX is not None else BEX_DEFAULT

        self.set_device_temperature(constants.T)

        # Setup switches
        self.SATLIM = math.exp(4.0)
        self.WMSI_factor = 10
        self.NR_damp_factor = options.nl_voltages_lock_factor

        sc, sc_reason = self._self_check()
        if not sc:
            raise Exception(sc_reason + " out of range")

    def set_device_temperature(self, T):
        """Change the temperature of the device.

        Correspondingly, ``VTO``, ``KP`` and ``PHI`` get updated.
        """
        self.TEMP = T
        self.VTO = self.VTO - self.TCV * (T - self.TNOM)
        self.KP = self.KP * (T / self.TNOM) ** self.BEX
        self.PHI = self.PHI * T / self.TNOM + 3.0 * constants.Vth(self.TNOM) * math.log(T / self.TNOM) \
            - constants.si.Eg(
                self.TNOM) * T / self.TNOM + constants.si.Eg(T)

    def get_device_temperature(self):
        """Returns the temperature of the device - in K.
        """
        return self.TEMP

    def print_model(self):
        """All the internal parameters of the model get printed out,
        for visual inspection. Notice some can be set to None
        (ie not available) if they were not provided in the netlist
        or some not provided are calculated from the others.
        """
        arr = []
        TYPE = 'N' if self.NPMOS == 1 else "P"
        arr.append(
            [self.name, "", "", TYPE + " MOS", "EKV MODEL", "", "", "", "",  "", "", ""])
        arr.append(["KP", "[A/V^2]", self.KP, "VTO", "[V]:", self.VTO,
                   "TOX", "[m]", self.TOX, "COX", "[F/m^2]:", self.COX])
        arr.append(["PHI", "[V]:", self.PHI, "GAMMA", "sqrt(V)", self.GAMMA,
                   "NSUB", "[cm^-3]", self.NSUB,  "VFB", "[V]:", self.VFB])
        arr.append(
            ["U0", "[cm^2/(V*s)]:", self.U0, "TCV", "[V/K]", self.TCV, "BEX", "", self.BEX,  "", "", ""])
        arr.append(["INTERNAL", "", "", "SAT LIMIT", "", self.SATLIM,
                   "W/M/S INV FACTOR", "", self.WMSI_factor,  "", "", ""])
        print(printing.table(arr))

    def get_voltages(self, vd, vg, vs):
        """Performs the VD <-> VS swap if needed.
        Returns:
        (VD, VG, VS) after the swap
        CS, an integer which equals to:
            +1 if no swap was necessary,
            -1 if VD and VS have been swapped.
        """
        # vd / vs swap
        vd = vd * self.NPMOS
        vg = vg * self.NPMOS
        vs = vs * self.NPMOS
        if vs > vd:
            vd_new = vs
            vs_new = vd
            cs = -1
        else:
            vd_new = vd
            vs_new = vs
            cs = +1

        return ((float(vd_new), float(vg), float(vs_new)), cs)

    def get_ip_abs_err(self, device):
        """Absolute error to be enforced in the calculation of the normalized currents.
        """
        return options.iea / (2.0 * constants.Vth(self.TEMP) ** 2.0 * self.KP * device.M * device.W / device.L)

    def setup_scaling(self, nq, device):
        """Calculates and stores in self.scaling the following factors:
          Ut, the thermal voltage,
          Is, the specific current,
          Gs, the specific transconductance,
          Qs, the specific charge.
        """
        self.scaling.Ut = constants.Vth()
        self.scaling.Is = 2 * nq * \
            self.scaling.Ut ** 2 * self.KP * device.W / device.L
        self.scaling.Gs = 2 * nq * \
            self.scaling.Ut * self.KP * device.W / device.L
        self.scaling.Qs = 2 * nq * self.scaling.Ut * self.COX
        return

    def get_vp_nv_nq(self, VG):
        """Calculates and returns:
            VP, the pinch-off voltage,
            nv, the slope factor,
            nq, the charge linearization factor.
        """
        VGeff = VG - self.VTO + self.PHI + self.GAMMA * math.sqrt(self.PHI)
        if VGeff > 0 and VG - self.VTO + (math.sqrt(self.PHI) + self.GAMMA / 2.0) ** 2 > 0:
            VP = VG - self.VTO - self.GAMMA * \
                (math.sqrt(VG - self.VTO + (math.sqrt(self.PHI) + self.GAMMA / 2.0) ** 2)
                 - (math.sqrt(self.PHI) + self.GAMMA / 2.0))
            if math.isnan(VP):
                VP = 0  # the argument of sqrt ^^ went negative
        else:
            VP = -self.PHI
        # print "VG", VG, "VGeff", VGeff, "VP", VP, self.GAMMA, self.PHI,
        # math.sqrt(VG -self.VTO +(math.sqrt(self.PHI)+self.GAMMA/2)**2), VG
        # -self.VTO +(math.sqrt(self.PHI)+self.GAMMA/2)**2
        nq = 1.0 + .5 * self.GAMMA / math.sqrt(self.PHI + .5 * VP)
        nv = 1.0 + .5 * self.GAMMA / math.sqrt(self.PHI + VP + 1e-12)

        return VP, nv, nq

    def get_ids(self, device, xxx_todo_changeme, opdict=None, debug=False):
        """Returns:
            IDS, the drain-to-source current (de-normalized),
            qs, the (scaled) charge at the source,
            qr, the (scaled) charge at the drain.
        """
        (vd, vg, vs) = xxx_todo_changeme
        if debug:
            print("=== Current for vd:", vd, "vg:", vg, "vs:", vs)
        ip_abs_err = self.get_ip_abs_err(device) if opdict[
            'ip_abs_err'] is None else opdict['ip_abs_err']

        (VD, VG, VS), CS_FACTOR = self.get_voltages(vd, vg, vs)

        # Weff, Leff = self.get_eff_wl(device.W, device.L)

        VP, nv, nq = self.get_vp_nv_nq(VG)

        self.setup_scaling(nq, device)

        vp = VP / self.scaling.Ut
        vs = VS / self.scaling.Ut
        vd = VD / self.scaling.Ut

        if debug:
            print("Scaled voltages: vd:", vd, "vp:", vp, "vs:", vs)

        v_ifn = vp - vs
        ifn = self.get_ismall(v_ifn, opdict['ip_abs_err'], max(
            opdict['ifn'], ISMALL_GUESS_MIN), debug=debug)

        if False:
            Leff = device.L
            v_irn = vp - vd
        else:
            Leff, v_irn = self.get_leq_virp(
                device, (vd, vg, vs), VP, device.L, ifn)

        irn = self.get_ismall(v_irn, opdict['ip_abs_err'], max(
            opdict['irn'], ISMALL_GUESS_MIN), debug=debug)

        if debug:
            print("vd:", vd, "vg:", VG / self.scaling.Ut, "vs:", vs, "vds:", vd - vs)
            print("v_ifn:", v_ifn, "v_irn:", v_irn)
            print("ifn:", ifn, "irn:", irn)
            print("ip_abs_err:", ip_abs_err)
            print("Vth:", self.scaling.Ut)
            print("nv", nv, "Is", self.scaling.Is)
            print("Weff:", device.W, "Leff:", Leff)
            print("NPMOS:", self.NPMOS, "CS_FACTOR", CS_FACTOR)

        qf = self.ismall2qsmall(ifn)
        qr = self.ismall2qsmall(irn)

        Ids =  CS_FACTOR * self.NPMOS * device.L / \
            Leff * device.M * self.scaling.Is * (ifn - irn)

        vd_real = vd if CS_FACTOR == 1 else vs
        vs_real = vs if CS_FACTOR == 1 else vd

        opdict.update(
            {'state': (vd_real * self.NPMOS, vg * self.NPMOS, vs_real * self.NPMOS)})
        opdict.update({'Ids': Ids, "Weff": device.W, "Leff": Leff, 'Vp': VP})
        opdict.update({'ifn': ifn, "irn": irn, "nv": nv, "nq": nq,
                      'beta': .5 * self.KP * device.W / Leff, 'Ispec': self.scaling.Is})
        opdict.update(
            {'VTH': self.VTO, "VOD": self.NPMOS * nv * (VP - VS), 'SAT': ifn > irn * self.SATLIM})
        opdict.update({'qf': qf * self.scaling.Qs, 'qr': qr * self.scaling.Qs})

        if max(ifn, irn) > self.WMSI_factor:
            WMSI = 2
        elif max(ifn, irn) < 1. / self.WMSI_factor:
            WMSI = 0
        else:
            WMSI = 1
        opdict.update({'WMSI': WMSI})

        if debug:
            print("current:", Ids)

        return Ids, qf, qr

    def get_leq_virp(self, device, xxx_todo_changeme1, Vp, Leff, ifn):
        # if ifn > 0 and Vp - constants.Vth()*vd > 0:
        (vd, vg, vs) = xxx_todo_changeme1
        assert vd >= vs
        Vc = self.UCRIT * device.N * Leff
        Vdss  = Vc * \
            (math.sqrt(.25 + constants.Vth() / Vc * math.sqrt(ifn)) - .5)
             # eq. 46
        # Drain-to-source saturation voltage for reverse normalized current,
        # eq. 47
        Vdssp = Vc * (math.sqrt(.25 + constants.Vth() / Vc * (math.sqrt(ifn) - .75 * math.log(ifn))) - .5) + \
            constants.Vth() * (math.log(.5 * Vc / constants.Vth()) - .6)

        # channel length modulation
        vser_1 = math.sqrt(ifn) - Vdss / constants.Vth()
        # if vser_1 < 0:
        #   vser_1 = 0
        Vds = (vd - vs) * .5 * constants.Vth()
        delta_v = 4.0 * constants.Vth() * math.sqrt(
            self.LAMBDA * vser_1 + 1.0 / 64)  # eq. 48
        Vip = math.sqrt(Vdss ** 2 + delta_v ** 2) - math.sqrt(
            (Vds - Vdss) ** 2 + delta_v ** 2)  # eq 50
        Lc = math.sqrt(constants.si.esi * self.XJ / self.COX)  # eq. 51
        delta_l = self.LAMBDA * Lc * \
            math.log(1 + (Vds - Vip) / (Lc * self.UCRIT))  # eq. 52

        # Equivalent channel length including channel-length modulation and
        # velocity saturation
        Lp = device.N * Leff - delta_l + (Vds + Vip) / self.UCRIT  # eq. 53
        Lmin = device.N * Leff / 10.0  # eq. 54
        Leq = .5 * (Lp + math.sqrt(Lp ** 2 + Lmin ** 2))  # eq. 55

        assert not math.isnan(Vdssp)
        assert not math.isnan(delta_v)

        v_irp = (Vp - Vds - vs * constants.Vth() - math.sqrt(Vdssp ** 2 + delta_v ** 2)
                 + math.sqrt((Vds - Vdssp) ** 2 + delta_v ** 2)) / constants.Vth()
        # else:
        #   v_irp = Vp/constants.Vth() - vd
        #   Leq = Leff

        return Leq, v_irp

    def get_gms(self, device, xxx_todo_changeme2, opdict=None, debug=False):
        """Returns the source-bulk transconductance or d(IDS)/d(VS-VB)."""
        (vd, vg, vs) = xxx_todo_changeme2
        (j1, j2, j3), CS_FACTOR = self.get_voltages(vd, vg, vs)
        Ids, qf, qr = self.get_ids(device, (vd, vg, vs), opdict, debug)
        if CS_FACTOR == +1:
            gms = -1.0 * self.scaling.Gs * qf
        elif CS_FACTOR == -1:
            gms = -self.scaling.Gs * qr
        return gms

    def get_gmd(self, device, xxx_todo_changeme3, opdict=None, debug=False):
        """Returns the drain-bulk transconductance or d(IDS)/d(VD-VB)."""
        (vd, vg, vs) = xxx_todo_changeme3
        (j1, j2, j3), CS_FACTOR = self.get_voltages(vd, vg, vs)
        Ids, qf, qr = self.get_ids(device, (vd, vg, vs), opdict, debug)
        if CS_FACTOR == +1:
            gmd = self.scaling.Gs * qr
        elif CS_FACTOR == -1:
            gmd = self.scaling.Gs * qf
        return gmd

    def get_gmg(self, device, xxx_todo_changeme4, opdict=None, debug=False):
        """Returns the gate-bulk transconductance or d(IDS)/d(VG-VB)."""
        (vd, vg, vs) = xxx_todo_changeme4
        VP, nv, nq = self.get_vp_nv_nq(float(vg))
        Ids, qf, qr = self.get_ids(device, (vd, vg, vs), opdict, debug)
        (j1, j2, j3), CS_FACTOR = self.get_voltages(vd, vg, vs)
        gmg = CS_FACTOR * self.scaling.Gs * (qf - qr) / nv
        return gmg

    def get_ismall(self, vsmall, ip_abs_err, iguess=None, debug=False):
        """Solves the problem: given v, find i such that:

        .. math::
            v = ln(q) + 2q

        ..math::
            q = sqrt(.25 + i) - .5

        The Newton Method is used inside.
        """
        # starting guess for Newton's Method.
        if iguess is None:
            iguess = 1.0
        # sanity checks
        if math.isnan(vsmall):
            raise ValueError("Attempted to calculate a current corresponding to a NaN voltage.")
        if not ip_abs_err > 0:
            raise ValueError("The normalized current absolute error has been set to a negative value.")
        ismall = newton(self._vsmall_obj, iguess, fprime=self._vsmall_obj_prime,
                        fprime2=self._vsmall_obj_prime2, args=(vsmall,),
                        tol=1.48e-8, maxiter=500)
        #print(ismall, max(ismall, 0))
        return max(ismall, 10*utilities.EPS)

    def _vsmall_obj(self, x, vsmall):
        """Returns :math:`e` according to the equations:
            q = sqrt(.25 + x) - .5
            e = ln(q) + 2q - vsmall
        """
        if x > 0:
            return math.log(math.sqrt(.25 + x) - 0.5) + 2.0 * math.sqrt(.25 + x) - 1.0 - vsmall
        else:
            return x - vsmall

    def _vsmall_obj_prime(self, x, vsmall):
        """The Newton algorithm in get_ismall(...) requires the evaluation of the
        first derivative of the fixed point function:
            dv/di = 1.0/(sqrt(.25+i)-.5) * .5/sqrt(.25 + i) + 1/sqrt(.25 + i)
        This is provided by this module.
        """
        if x < utilities.EPS:
            x = 10*utilities.EPS
        return 0.5/(math.sqrt(.25 + x) - .5)/math.sqrt(.25 + x) + \
               1.0/math.sqrt(.25 + x)

    def _vsmall_obj_prime2(self, x, vsmall):
        if x < utilities.EPS:
            x = 10*utilities.EPS
        return -1./(4*(x + 0.25)*(math.sqrt(x + 0.25) - 0.5)**2) - 1./(2.*(x + 0.25)**(3./2.)) - 1./(4.*(x + 0.25)**(3./2.)*(math.sqrt(x + 0.25) - 0.5))

    def get_vsmall(self, ismall, verbose=3):
        """Returns v according to the equations:
            q = sqrt(.25 + i) - .5
            v = ln(q) + 2q
        """
        if abs(ismall) < utilities.EPS:
            ismall = utilities.EPS  # otherwise we get log(0)
            if verbose == 6:
                print("EKV: Machine precision limited the resolution on i. (i<EPS)")
        vsmall = math.log(math.sqrt(.25 + ismall) - 0.5) + \
                 2.0 * math.sqrt(.25 + ismall) - 1.0
        return vsmall

    def get_dvsmall_dismall(self, ismall, verbose=3):
        """The Newton algorithm in get_ismall(...) requires the evaluation of the
        first derivative of the fixed point function:
            dv/di = 1.0/(sqrt(.25+i)-.5) * .5/sqrt(.25 + i) + 1/sqrt(.25 + i)
        This is provided by this module.
        """
        if abs(ismall) < utilities.EPS:
            ismall = utilities.EPS
            if verbose == 6:
                print("EKV: Machine precision limited the resolution on dv/di in the NR iteration. (i<EPS)")
        dvdi = 1.0 / (math.sqrt(.25 + ismall) - .5) * .5 / \
            math.sqrt(.25 + ismall) + 1.0 / math.sqrt(.25 + ismall)
        return dvdi

    def ismall2qsmall(self, ismall, verbose=0):
        """ i(f,r) -> q(f,r)
        Convert a source/drain scaled current to the corresponding normalized charge."""
        if verbose == 6:  # ismall is lower than EPS, errors here are usually not important
            print("EKV: Machine precision limited the resolution on q(s,d). (i<EPS)")
        ismall = max(0, ismall)
        qsmall = math.sqrt(.25 + ismall) - .5
        return qsmall

    def qsmall2ismall(self, qsmall):
        """ q(f,r) -> i(f,r)
        Convert a source/drain scaled charge to the corresponding normalized current."""
        ismall = qsmall ** 2 + qsmall
        return ismall

    def _self_check(self):
        """Performs sanity check on the model parameters."""
        ret = True, ""
        if self.NSUB is not None and self.NSUB < 0:
            ret = (False, "NSUB " + str(self.NSUB))
        elif self.U0 is not None and not self.U0 > 0:
            ret = (False, "UO " + str(self.U0))
        elif not self.GAMMA > 0:
            ret = (False, "GAMMA " + str(self.GAMMA))
        elif not self.PHI > 0.1:
            ret = (False, "PHI " + str(self.PHI))
        return ret

    def _device_check(self, adev):
        """Performs sanity check on the device parameters."""
        if not adev.L > 0:
            ret = (False, "L")
        elif not adev.W > 0:
            ret = (False, "W")
        elif not adev.N > 0:
            ret = (False, "N")
        elif not adev.M > 0:
            ret = (False, "M")
        else:
            ret = (True, "")
        return ret




# -*- coding: iso-8859-1 -*-
# mosq.py
# Implementation of the square-law MOS transistor model
# Copyright 2012 Giuseppe Venturini
#
# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

"""
The Square Law Mos Model
------------------------

This module defines two classes:

- :class:`mosq_device`, the device
- :class:`mosq_model`, the model


Implementation details
----------------------

Assuming :math:`V_{ds} > 0` and a transistor type N in the
following, we have the following regions implemented:

1. No subthreshold conduction.
       - :math:`V_{gs} < V_T`
       - :math:`I_D = 0`
2. Ohmic region
       - :math:`V_{GS} > V_T` and :math:`V_{GD} > V_T`
       - :math:`I_D = k_n W/L ((V_{GS}-V_{T})V_{DS} - V_{DS}^2/2)`
3. Saturation region
       - :math:`V_{GS} > V_T` and :math:`V_{DS} > V_{GS} - V_{T}`
       - :math:`V_{GS} < V_{T}`
       - :math:`I_D = 1/2 k_n W/L (V_{GS}-V_T)^2 * [1 + \lambda*(V_{DS}-V_{GS}+V_T)]`

Module reference
----------------

"""
from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import math
import numpy as np

from . import constants
from . import options
from . import utilities
from . import printing

# DEFAULT VALUES FOR 500n CH LENGTH
COX_DEFAULT = .7e-3
VTO_DEFAULT = .5
GAMMA_DEFAULT = 1.0
PHI_DEFAULT = .7
KP_DEFAULT = 50e-6
LAMBDA_DEFAULT = .5
AVT_DEFAULT = 7.1e-3 * 1e-6
AKP_DEFAULT = 1.8e-2 * 1e-6

TCV_DEFAULT = 1e-3
BEX_DEFAULT = -1.5

ISMALL_GUESS_MIN = 1e-10


class mosq_device(object):
    def __init__(self, part_id, nd, ng, ns, nb, W, L, model, M=1, N=1):
        """Quadratic Law MOSFET device

        **Parameters:**

        part_id : string
            The part ID of the model. Eg. ``'M1'`` or ``'Mlow'``, the first
            letter should always be ``'M'``.
        nd : int
            drain node
        ng : int
            gate node
        ns : int
            source node
        nb : int
            bulk node
        L : float
            element width [m]
        W : float
            element length [m]
        model : mosq_mos_model instance
            the model for the device
        M : int, optional
            shunt multiplier (n. of shunt devices)
        N : int, optional
            series multiplier (n. of series devices)

        """
        self.ng = ng
        self.nb = nb
        self.n1 = nd
        self.n2 = ns
        self.ports = ((self.n1, self.n2), (
            self.ng, self.n2), (self.nb, self.n2))

        class dev_class(object):
            pass  # empty class to hold device parameters
        self.device = dev_class()
        self.device.L = float(L)  # channel length -
        self.device.W = float(W)  # channel width -
        self.device.M = int(M)  # parallel multiple device number
        self.device.N = int(N)  # series multiple device number
        self.device.mckey = None
        self.device.part_id = part_id
        self.mosq_model = model
        self.mc_enabled = False
        self.opdict = {}
        self.opdict.update( {'state':(float('nan'), float('nan'),
                                      float('nan'))})
        self.part_id = part_id
        self.is_nonlinear = True
        self.is_symbolic = True
        self.dc_guess = [self.mosq_model.VTO*0.4*self.mosq_model.NPMOS,
                         self.mosq_model.VTO*1.1*self.mosq_model.NPMOS,
                         0]

        devcheck, reason = self.mosq_model.device_check(self.device)
        if not devcheck:
            raise ValueError(reason + " out of boundaries.")

    def get_drive_ports(self, op):
        """Get the circuit ports that drive the device.

        **Returns:**

        tp : a tuple of tuples of nodes, each node being a drive port of the
        device.

        Eg. ``tp`` might be defined as:

        ::

            tp = (port0, port1, port2...)

        Where each port in the tuple is of the form:

        ::

            port0 = (nplus, nminus)

        In the case of a MOSQ device, the method returns:

        ::

            ((nd, nb), (ng, nb), (ns, nb))

        Where:

        * ``nd`` is the internal identifier of the drain node,
        * ``ng`` is the internal identifier of the gate node,
        * ``ns`` is the internal identifier of the source node.
        * ``nb`` is the internal identifier of the bulk node,

        """
        return self.ports  # d,g,b

    def get_output_ports(self):
        """Get the circuit ports where the device injects current.

        **Returns:**

        ports : a tuple of tuples of nodes, such as as:

        ::

            (port0, port1, port2...)

        Where each port in the tuple is itself a tuple, made of two nodes, eg.

        ::

            port0 = (nplus, nminus)

        In the case of a MOS device, the method returns:

        ::

            ((nd, ns),)

        Where:

        * ``nd`` is the internal identifier of the drain node,
        * ``ns`` is the internal identifier of the source node.

        """
        return ((self.n1, self.n2),)

    def __str__(self):
        rep = self.part_id + " %(nd)s %(ng)s %(ns)s %(nb)%s " + \
              self.mosq_model.name + " w=" + str(self.device.W) + " l=" + \
              str(self.device.L) + " M=" + str(self.device.M) + " N=" + \
              str(self.device.N)
        return rep

    def _get_mos_type(self):
        """Returns N or P (capitalized), depending on the device type.
        """
        mtype = 'N' if self.mosq_model.NPMOS == 1 else 'P'
        return mtype

    def istamp(self, ports_v, time=0, reduced=True):
        """Get the current stamp matrix

        A stamp matrix corresponding to the current flowing in the element
        with the voltages applied as specified in the ``ports_v`` vector.

        **Parameters:**

        ports_v : list
            A list in the form: ``[voltage_across_port0, voltage_across_port1,
            ...]``
        time: float
            the simulation time at which the evaluation is performed.
            It has no effect here. Set it to ``None`` during DC analysis.

        """
        sw_vect, CS = self.mosq_model.get_voltages(*ports_v)
        ids = self.mosq_model.get_ids(self.mosq_model, self.device, sw_vect)
        istamp = np.array((CS*ids, -CS*ids), dtype=np.float64)
        indices = ((self.n1 - 1*reduced, self.n2 - 1*reduced), (0, 0))
        if reduced:
            delete_i = [pos for pos, i in enumerate(indices[0]) if i == -1]
            istamp = np.delete(istamp, delete_i, axis=0)
            indices = tuple(zip(*[(i, j) for i, j in zip(*indices) if i != -1]))
        return indices, istamp

    def update_status_dictionary(self, ports_v):
        """Update the status dictionary

        The status dictionary may be accessed at ``elem.opdict`` and contains
        several pieces of information that may be of interest regarding the
        biasing of the MOS device.
        """
        if self.opdict is None:
            self.opdict = {}
        if not (self.opdict['state'] == ports_v[0]).all() or 'gmd' not in self.opdict \
            or 'gm' not in self.opdict or 'gmb' not in self.opdict \
            or 'Ids' not in self.opdict or 'SAT' not in self.opdict:

            vds, vgs, _ = ports_v[0]
            self.opdict['state'] = ports_v[0]
            gstamp = self.gstamp(ports_v[0], reduced=False)[1]
            self.opdict['gmd'] = gstamp[0, 0]
            self.opdict['gm'] = gstamp[0, 1]
            self.opdict['gmb'] = gstamp[0, 3]
            self.opdict['Ids'] = self.istamp(ports_v[0], reduced=False)[1][0]
            self.opdict.update({'VTH':self.mosq_model.get_VT(ports_v[0], self.device)})
            self.opdict.update({'W':self.device.W, 'L':self.device.L,
                                'ON':(vgs >= self.opdict['VTH'])})
            self.opdict.update({'beta':.5*self.mosq_model.KP*self.device.W/self.device.L})
            self.opdict.update({'VOD':self.mosq_model.NPMOS*(vgs - self.opdict['VTH']),
                                'SAT':vds > vgs - self.opdict['VTH']})
        else:
            pass
            #already up to date

    def get_op_info(self, ports_v):
        """Information regarding the Operating Point (OP)

        **Parameters:**

        ports_v : list of lists
            The voltages applied to all the driving ports, grouped by output
            port.

        i.e.

        ::

            [<list of voltages for the drive ports of output port 0>,
             <list of voltages for the drive ports of output port 1>,
             ...,
             <list of voltages for the drive ports of output port N>]

        Usually, this method returns ``op_keys`` and the corresponding
        ``op_info``, two lists, one holding the labels, the other the
        corresponding values.

        In the case of MOSFETs, the values are way too many to be shown in a
        linear table. For this reason, we return ``None`` as ``op_keys``, and we
        return for ``op_info`` a list which holds both labels and values in a
        table-like manner, spanning the vertical and horizontal dimension.

        For this reason, each MOSFET has to have its OP info printed alone, not
        grouped as it happens with most other elements.

        **Returns:**

        op_keys : ``None``
            See above for why this value is always ``None``.
        op_info : list of floats
            The OP information ready to be passed to :func:`printing.table` for
            arranging it in a pretty table to display.
        """
        self.update_status_dictionary(ports_v)
        sat_status = "SATURATION" if self.opdict['SAT'] else "LINEAR"
        if not self.opdict["ON"]:
            status = "OFF"
        else:
            status = "ON"

        arr = [[self.part_id + " ch", status, "", "", sat_status, "", "", "",
                "", "", "", ""], ]
        arr.append(["beta", "[A/V^2]:", self.opdict['beta'], "Weff", "[m]:",
                    str(self.opdict['W']) + " (" + str(self.device.W) + ")",
                    "L", "[m]:", str(self.opdict['L']) + " (" +
                    str(self.device.L) + ")", "M/N:", "", str(self.device.M) +
                    "/" + str(self.device.N)])
        arr.append(["Vds", "[V]:", float(ports_v[0][0]), "Vgs", "[V]:",
                    float(ports_v[0][1]), "Vbs", "[V]:", float(ports_v[0][2]),  "",
                    "", ""])
        arr.append(["VTH", "[V]:", self.opdict['VTH'], "VOD", "[V]:",
                    self.opdict['VOD'], "", "", "", "VA", "[V]:",
                    str(self.opdict['Ids'] / self.opdict['gmd'])])
        arr.append(["Ids", "[A]:", self.opdict['Ids'], "", "", "", "", "", "",
                    "", "", ''])
        arr.append(["gm", "[S]:", self.opdict['gm'], "gmb", "[S]:",
                    self.opdict['gmb'], "ro", u"[\u2126]:",
                    1./self.opdict['gmd'], "", "", ""])

        return None, arr

    def gstamp(self, ports_v, time=0, reduced=True):
        """Get the transconductance stamp matrix

        **Parameters:**

        ports_v : sequence
            a sequence of the form: ``[voltage_across_port0,
            voltage_across_port1, ...]``
        time : float, optional
            the simulation time at which the evaluation is performed. Set it to
            ``None`` during DC analysis. Defaults to 0.
        reduced : bool, optional
            Whether the returned matrix should be in reduced form or not.
            Defaults to ``True``, corresponding to reduced form.

        **Returns:**

        indices : sequence of sequences
            The indices corresponding to the stamp matrix.
        stamp : ndarray
            The stamp matrix.
        """
        indices = ([self.n1 - 1]*4 + [self.ng - 1]*4 + [self.n2 - 1]*4 + [self.nb - 1]*4,
                   [self.n1 - 1, self.ng - 1, self.n2 - 1, self.nb - 1]*4)

        sw_vect, CS = self.mosq_model.get_voltages(*ports_v)
        gmd = self.mosq_model.get_gmd(self.mosq_model, self.device, sw_vect)
        gmg = self.mosq_model.get_gm(self.mosq_model, self.device, sw_vect)
        gmb = self.mosq_model.get_gmb(self.mosq_model, self.device, sw_vect)
        if gmd == 0:
            gmd = options.gmin*2
        if gmg == 0:
            gmg = options.gmin*2
        if gmb == 0:
            gmb = -2*options.gmin
        stamp = np.array(((gmd, gmg, -gmd-gmb-gmg, gmb),
                          (0, 0, 0, 0),
                          (-gmd, -gmg, gmd + gmg + gmb, -gmb),
                          (0, 0, 0, 0)), dtype=np.float64)
        if CS == -1:
            stamp = self.mosq_model.T1*stamp*self.mosq_model.T2
        if (self.opdict['state'] != ports_v[0]).any():
            self.opdict = {'state':ports_v[0]}
        self.opdict.update({'gmd': stamp[0, 0]})
        self.opdict.update({'gm': stamp[0, 1]})
        self.opdict.update({'gmb': stamp[0, 3]})
        if reduced:
            zap_rc = [pos for pos, i in enumerate(indices[1][:4]) if i == -1]
            stamp = np.delete(stamp, zap_rc, axis=0)
            stamp = np.delete(stamp, zap_rc, axis=1)
            indices = tuple(zip(*[(i, y) for i, y in zip(*indices) if (i != -1 and y != -1)]))
            stamp_flat = stamp.reshape(-1)
            stamp_folded = []
            indices_folded = []
            for ix, it in enumerate([(i, y) for i, y in zip(*indices)]):
                if it not in indices_folded:
                    indices_folded.append(it)
                    stamp_folded.append(stamp_flat[ix])
                else:
                    w = indices_folded.index(it)
                    stamp_folded[w] += stamp_flat[ix]
            indices = tuple(zip(*indices_folded))
            stamp = np.array(stamp_folded)
        return indices, stamp

    def get_value_function(self, identifier):
        def get_value(self):
            return self.opdict[identifier]
        return get_value

    def get_mc_requirements(self):
        return True, 2

    def setup_mc(self, status, mckey):
        self.mc_enabled = status
        if self.mc_enabled:
            self.device.mckey = mckey
        else:
            self.device.mckey = None

    def get_netlist_elem_line(self, nodes_dict):
        """Get the element netlist entry"""
        mos_type = self._get_mos_type()
        return "%s %s %s %s %s %s type=%s w=%g l=%g m=%g n=%g" % \
              (self.part_id, nodes_dict[self.n1], nodes_dict[self.ng],
              nodes_dict[self.n2], nodes_dict[self.nb], self.mosq_model.name,
              mos_type, self.device.W, self.device.L, self.device.M,
              self.device.N)


class mosq_mos_model(object):
    def __init__(self, name=None, TYPE='n', TNOM=None, COX=None,
                 GAMMA=None, NSUB=None, PHI=None, VTO=None, KP=None,
                 LAMBDA=None, AKP=None, AVT=None,
                 TOX=None, VFB=None, U0=None, TCV=None, BEX=None):

        self.name = "model_mosq0" if name is None else name
        self.TNOM = float(TNOM) if TNOM is not None else constants.Tref
        # print "TYPE IS:" + TYPE
        if TYPE.lower() == 'n':
            self.NPMOS = 1
        elif TYPE.lower() == 'p':
            self.NPMOS = -1
        else:
            raise ValueError("Unknown MOS type %s" % TYPE)

        # optional parameters (no defaults)
        self.TOX = float(TOX) if TOX is not None else None
        self.NSUB = float(NSUB) if NSUB is not None else None
        self.VFB = self.NPMOS * float(VFB) if VFB is not None else None
        self.U0 = float(U0) if U0 is not None else None

        # crucial parameters
        if COX is not None:
            self.COX = float(COX)
        elif TOX is not None:
            self.COX = constants.si.eox / TOX
        else:
            self.COX = COX_DEFAULT

        if GAMMA is not None:
            self.GAMMA = float(GAMMA)
        elif NSUB is not None:
            self.GAMMA = math.sqrt(
                2 * constants.e * constants.si.esi * NSUB * 10 ** 6 / self.COX)
        else:
            self.GAMMA = GAMMA_DEFAULT
        if PHI is not None:
            self.PHI = float(PHI)
        elif NSUB is not None:
            self.PHI = 2 * constants.Vth(self.TNOM) * math.log(
                NSUB * 10 ** 6 / constants.si.ni(self.TNOM))
        else:
            self.PHI = PHI_DEFAULT
        if VTO is not None:
            self.VTO = self.NPMOS * float(VTO)
            if self.VTO < 0:
                print("(W): model %s has internal negative VTO (%f V)." % (self.name, self.VTO))
        elif VFB is not None:
            self.VTO = VFB + PHI + GAMMA * PHI  # inv here??
        else:
            self.VTO = VTO_DEFAULT

        if KP is not None:
            self.KP = float(KP)
        elif U0 is not None:
            self.KP = (U0 * 10 ** -4) * self.COX
        else:
            self.KP = KP_DEFAULT

        self.LAMBDA = LAMBDA if LAMBDA is not None else LAMBDA_DEFAULT
        # Intrinsic model temperature parameters
        self.TCV = self.NPMOS * \
            float(TCV) if TCV is not None else self.NPMOS * TCV_DEFAULT
        self.BEX = float(BEX) if BEX is not None else BEX_DEFAULT

        # Monte carlo
        self.AVT = AVT if AVT is not None else AVT_DEFAULT
        self.AKP = AKP if AKP is not None else AKP_DEFAULT

        self.set_device_temperature(constants.T)

        sc, sc_reason = self._self_check()
        if not sc:
            raise Exception(sc_reason + " out of range")
        self.T1 = np.array(((0, 0, 1, 0),
                            (0, 1, 0, 0),
                            (1, 0, 0, 0),
                            (0, 0, 0, 1)))
        self.T2 = np.array(((0, 0, 1, 0),
                            (0, 1, 0, 0),
                            (1, 0, 0, 0),
                            (0, 0, 0, 1)))

    def set_device_temperature(self, T):
        """Change the temperature of the device.

        Correspondingly, ``VTO``, ``KP`` and ``PHI`` get updated.
        """
        self.TEMP = T
        self.VTO = self.VTO - self.TCV * (T - self.TNOM)
        self.KP = self.KP * (T / self.TNOM) ** self.BEX
        self.PHI = (self.PHI * T / self.TNOM + 3 * constants.Vth(self.TNOM) *
                    math.log(T / self.TNOM) - constants.si.Eg(self.TNOM) * T /
                    self.TNOM + constants.si.Eg(T))

    def get_device_temperature(self):
        """Returns the temperature of the device - in K.
        """
        return self.TEMP

    def print_model(self):
        """Print out the model

        All the internal parameters of the model get printed out, for visual
        inspection. Notice some can be set to ``None`` (ie not available) if
        they were not provided and some of those not provided are calculated
        from the others.
        """
        arr = []
        TYPE = 'N' if self.NPMOS == 1 else "P"
        arr.append([self.name, "", "", TYPE + " MOS", "SQUARE MODEL", "", "",
                    "", "",  "", "", ""])
        arr.append(["KP", "[A/V^2]", self.KP, "VTO", "[V]:", self.VTO,
                    "TOX", "[m]", self.TOX, "COX", "[F/m^2]:", self.COX])
        arr.append(["PHI", "[V]:", self.PHI, "GAMMA", "sqrt(V)", self.GAMMA,
                    "NSUB", "[cm^-3]", self.NSUB,  "VFB", "[V]:", self.VFB])
        arr.append(["U0", "[cm^2/(V*s)]:", self.U0, "TCV", "[V/K]", self.TCV,
                    "BEX", "", self.BEX,  "", "", ""])
        print(printing.table(arr))

    def get_voltages(self, vds, vgs, vbs):
        """Performs the D <-> S swap if needed.

        **Returns:**

        voltages : tuple
            A tuple containing ``(VDS, VGS, VBS)`` after the swap
        CS : int
            ``CS`` is an integer which equals to:
            * +1 if no swap was necessary,
            * -1 if VD and VS have been swapped.
        """
        # vd / vs swap
        vds = float(vds)
        vgs = float(vgs)
        vbs = float(vbs)
        vds = vds * self.NPMOS
        vgs = vgs * self.NPMOS
        vbs = vbs * self.NPMOS
        if vds < 0:
            vds_new = -vds
            vgs_new = vgs - vds
            vbs_new = vbs - vds
            cs = -1
        else:
            vds_new = vds
            vgs_new = vgs
            vbs_new = vbs
            cs = +1
        # print ((float(vds_new), float(vgs_new), float(vbs_new)), cs)
        return (float(vds_new), float(vgs_new), float(vbs_new)), cs

    def get_svt_skp(self, device, debug=False):
        if device.mckey and debug:
            print("Monte carlo enabled. key:", device.mckey)
        if device.mckey:
            svt = device.mckey[0] * self.AVT / math.sqrt(
                2 * device.W * device.L)
            skp = device.mckey[1] * self.AKP / math.sqrt(
                2 * device.W * device.L)
        else:
            svt, skp = 0, 0
        return svt, skp

    def get_VT(self, voltages, device):
        """Get the threshold voltage"""
        #vds, vgs, vbs = voltages
        _, _, vbs = voltages
        #(_, _, vbs), CS = self.get_voltages(*voltages)
        vsqrt1 = max(-vbs + 2*self.PHI, 0.)
        vsqrt2 = max(2*self.PHI, 0.)
        svt, _ = self.get_svt_skp(device)
        VT = self.VTO + svt + self.GAMMA * (math.sqrt(vsqrt1) -
                                            math.sqrt(vsqrt2))
        return VT

    @utilities.memoize
    def get_ids(self, device, voltages):
        """Get the drain-source current

        **Parameters:**

        device : object
            The device object holding the device parameters
            as attributes.
        voltages : tuple
            A tuple containing the voltages applied to the driving ports.
            In this case, the tuple is ``(vds, vgs, vbs)``.

        **Returns:**

        ids : float
            The drain-source current
        """
        """
        Returns:
            IDS, the drain-to-source current
        """
        (vds, vgs, vbs) = voltages
        debug = False
        if debug:
            print("=== %s (%sch) current for vds: %g, vgs: %g, vbs: %g" \
                  % (device.part_id, 'n'*(self.NPMOS == 1) +
                  'p'*(self.NPMOS == -1), vds, vgs, vbs))

        if debug:
            print("PHI:", self.PHI, "vbs:", vbs)

        VT = self.get_VT((vds, vgs, vbs), device)
        _, skp = self.get_svt_skp(device)
        if vgs < VT:
            ids = options.iea * (vgs / VT + vds / VT) / 100
            if debug:
                print("OFF: %g" % ids)
        else:
            if vds < vgs - VT -0.5*self.LAMBDA*(VT - vgs)**2:
                ids = (skp + 1) * self.KP * device.W / \
                      device.L * ((vgs - VT) * vds - .5 * vds ** 2)
                if debug:
                    print("OHMIC: %g" % ids)
            else:
                ids = (skp + 1) * .5 * self.KP * device.W / device.L * (
                      vgs - VT) ** 2 * (1 + self.LAMBDA * (vds - vgs + VT + 0.25*self.LAMBDA*(VT - vgs)**2))
                if debug:
                    print("SAT: %g" % ids)
        Ids = self.NPMOS * device.M / device.N * ids

        return Ids

    @utilities.memoize
    def get_gmb(self, device, voltages):
        """Get the bulk-source transconductance

        Mathematically:

        .. math::
            g_{mb} = \\frac{dI_{DS}}{d(VS-VB)}

        **Parameters:**

        device : object
            The device object holding the device parameters
            as attributes.
        voltages : tuple
            A tuple containing the voltages applied to the driving ports.
            In this case, the tuple is ``(vds, vgs, vbs)``.

        **Returns:**

        gmb : float
            The source-bulk transconductace.
        """
        (vds, vgs, vbs) = voltages
        debug = False
        svt, skp = self.get_svt_skp(device, debug=False)
        assert vds >= 0
        vsqrt1 = max(-vbs + 2*self.PHI, 0.)
        vsqrt2 = max(2*self.PHI, 0.)
        VT = self.VTO + svt + self.GAMMA * \
            (math.sqrt(vsqrt1) - math.sqrt(vsqrt2))
        gmb = 0
        if vgs < VT:
            pass # gmb = 0
        else:
            if vds < vgs - VT:
                if vsqrt1 > 0:
                    gmb = self.KP * self.GAMMA * vds * device.W / \
                          (2 * device.L * vsqrt1 ** .5)
            else:
                if vsqrt1 > 0:
                    gmb += -0.25*self.KP*self.GAMMA*self.LAMBDA*device.W * (vsqrt1 > 0) * \
                           (-self.GAMMA*(-vsqrt2**.5 + vsqrt1**.5) + vgs - self.VTO)**2 / \
                           (device.L * vsqrt1**.5)
                    gmb += +0.5*self.KP*self.GAMMA*device.W*(self.LAMBDA* \
                            (self.GAMMA * (vsqrt2**.5 + vsqrt1**.5) + vds - vgs + self.VTO) + 1.0) *\
                            (-self.GAMMA * (vsqrt2**.5 + vsqrt1**.5) \
                            + vgs - self.VTO) / (device.L * vsqrt1**.5)
        gmb = self.NPMOS * (1 + skp) * gmb * device.M / device.N
        if debug:
            print("gmb %g" % gmb)
        return gmb

    @utilities.memoize
    def get_gmd(self, device, voltages):
        """Get the drain-source transconductance

        Mathematically:

        .. math::
            g_{md} = \\frac{dI_{DS}}{d(VD-VS)}

        **Parameters:**

        device : object
            The device object holding the device parameters
            as attributes.
        voltages : tuple
            A tuple containing the voltages applied to the driving ports.
            In this case, the tuple is ``(vds, vgs, vbs)``.

        **Returns:**

        gmb : float
            The drain-source transconductace.
        """
        (vds, vgs, vbs) = voltages
        debug = False
        svt, skp = self.get_svt_skp(device, debug=False)
        assert vds >= 0
        vsqrt1 = max(-vbs + 2*self.PHI, 0.)
        vsqrt2 = max(2*self.PHI, 0.)
        VT = self.VTO + svt + self.GAMMA * \
            (math.sqrt(vsqrt1) - math.sqrt(vsqrt2))
        if vgs < VT:
            gmd = options.iea / VT / 100
        else:
            if vds < vgs -VT -0.5*self.LAMBDA*(VT - vgs)**2: # correction term disc. due to LAMBDA
                gmd = self.KP * device.W / device.L * (vgs - vds - VT)
            else:
                gmd = 0.5 * self.KP * self.LAMBDA * device.W / device.L * \
                      (vgs - VT)**2
        gmd = (1 + skp) * gmd * device.M / device.N
        if debug:
            print("gmd %g" % gmd)
        return gmd

    @utilities.memoize
    def get_gm(self, device, voltages):
        """Get the gate-source transconductance

        Mathematically:

        .. math::
            g_{ms} = \\frac{dI_{DS}}{d(VG-VS)}

        Often this is referred to as just :math:`g_m`.

        **Parameters:**

        device : object
            The device object holding the device parameters
            as attributes.
        voltages : tuple
            A tuple containing the voltages applied to the driving ports.
            In this case, the tuple is ``(vds, vgs, vbs)``.

        **Returns:**

        gmb : float
            The gate-source transconductace.
        """
        (vds, vgs, vbs) = voltages
        debug = False
        svt, skp = self.get_svt_skp(device, debug=False)
        assert vds >= 0
        vsqrt1 = max(-vbs + 2*self.PHI, 0.)
        vsqrt2 = max(2*self.PHI, 0.)
        VT = self.VTO + svt + self.GAMMA * \
            (math.sqrt(vsqrt1) - math.sqrt(vsqrt2))
        if vgs < VT:
            gm = options.iea / VT / 100
        else:
            if vds < vgs - VT:
                gm = self.KP * device.W / device.L * vds
            else:
                gm = -0.5*self.KP*self.LAMBDA * device.W/device.L * (-self.GAMMA*(-vsqrt2**.5 + vsqrt1**.5) + vgs - self.VTO)**2 \
                     +0.5*self.KP * device.W/device.L *(self.LAMBDA*( self.GAMMA*(-vsqrt2**.5 + vsqrt1**.5) + vds - vgs + self.VTO) + 1.0) *\
                     (-2 * self.GAMMA * (-vsqrt2**.5 + vsqrt1**.5) + 2*vgs - 2*self.VTO)
        gm = (1 + skp) * gm * device.M / device.N
        if debug:
            print("gmg %g" % gm)
        return gm

    def _self_check(self):
        """Performs sanity check on the model parameters."""
        ret = True, ""
        if self.NSUB is not None and self.NSUB < 0:
            ret = (False, "NSUB " + str(self.NSUB))
        elif self.U0 is not None and not self.U0 > 0:
            ret = (False, "UO " + str(self.U0))
        elif not self.GAMMA > 0:
            ret = (False, "GAMMA " + str(self.GAMMA))
        elif not self.PHI > 0.1:
            ret = (False, "PHI " + str(self.PHI))
        elif self.AVT and self.AVT < 0:
            ret = (False, "AVT " + str(self.AVT))
        elif self.AKP and self.AKP < 0:
            ret = (False, "AKP " + str(self.AKP))
        return ret

    def device_check(self, adev):
        """Performs sanity check on the device parameters."""
        if not adev.L > 0:
            ret = (False, "L")
        elif not adev.W > 0:
            ret = (False, "W")
        elif not adev.N > 0:
            ret = (False, "N")
        elif not adev.M > 0:
            ret = (False, "M")
        else:
            ret = (True, "")
        return ret




# -*- coding: iso-8859-1 -*-
# gear.py
# Gear Linear Multi-Step (LMS) DF
# Copyright 2006 Giuseppe Venturini

# This file is part of the ahkab simulator.
#
# Ahkab is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2 of the License.
#
# Ahkab is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License v2
# along with ahkab.  If not, see <http://www.gnu.org/licenses/>.

# This file should be conforming to a standard

""" About the method:
    This is an implicit method, it means that to compute dx(n+1)/dt the value of x in (n+1) is required.
    We don't know it, since it's our objective).
    This method, as all other implicit methods, allows us to write the derivative as:
     dx(n+1)/dt = x_coeff * x(n+1) + const                  (ii)

    The get_df method returns those two vectors.

    Gear's LMS interpolates the solution in a number of points equal to its order. Since it's a implicit
    method, one of these is x(n+1). The values x(n), x(n-1)... x(n-(order+2)) need to be supplied to the method.
    We can write x(t) as:
     x(t)   = a0 + a1*(t(n+1) -  t ) + a2*( t(n+1) - t )^2 + ...        (i)
    The equation has <order> a coefficients, which we need to determine.
    For this reason, we write a system of "order" equations in this way:
     x(n+1) = a0 + a1*(t(n+1) - t(n+1)) + a2*(t(n+1) - t(n+1))^2 + ...
     x(n)   = a0 + a1*(t(n+1) -  t(n) ) + a2*( t(n+1) - t(n) )^2 + ...
     x(n-1) = a0 + a1*(t(n+1) - t(n-1)) + a2*(t(n+1) - t(n-1))^2 + ...


    Which may be rewritten as:

      z = A * a

    z is the vector of known values of x
    A is a time dependant matrix
    a is a vector made of the a* coeffiecients

    We don't need to explicit ALL of the a* coeffiecients. What we are really looking for is the derivative of x
    in t(n+1), dx(n+1)/dt in short.
    If we differentiate the relation (i):
      dx(t)/dt = -a1 - 2*a2*( t(n+1) - t ) - 3*a3*( t(n+1) - t )^2 ...
    Which evaluated in t = t(n+1) gives:
      dx(n+1)/dt = -1 * a1

    Our objective is then a1.
    From the previsious system we write:
     a = A^-1 * z
    a1 is a[1,0], which may be extracted in this way:
     et = [0 1 0 0 0 0 ...] (order elements)
     a1 = et * a = et * A^-1 * z
    Because of the associative prperty of matrix multiplication, we can write:

     P  = et * A^-1
     a1 = P[1, :] * z

    But, we don't know z[0,0] = x(n+1), we can split the above relation:

     a1 = P[1, 0] * x(n+1) + P[1, 1:] * z[1:, 0]

    We arrived to the relation written above (ii)
     dx(n+1)/dt = x_coeff * x(n+1) + const = -1*a1

    So:
     x_coeff = -1 * P[1, 0]
     const   = -1 * P[1, 1:] * z[1:, 0]

    This module uses a faster way to compute the values that doesn't require to invert the matrix.
    Anyway, from a theorical point of view, the above applies.
      """

from __future__ import (unicode_literals, absolute_import,
                        division, print_function)

import numpy as np
import sys

from scipy.misc import factorial

from . import printing

order = None
# FAST = True


def is_implicit():
    return True


def has_ff():
    return True


def get_required_values():
    """This returns two python arrays built this way:
    [ max_order_of_x, max_order_of_dx ]
    Where:
    Both the values are int, or None
    if max_order_of_x is set to k, the df method needs all the x(n-i) values of x,
    where i<=k (the value the function assumed i+1 steps before the one we will ask for the derivative).
    The same applies to max_order_of_dx, but regards dx(n)/dt
    None means that NO value is required.

    The first array has to be used if no prediction is required, the second are the values needed for prediction.
    """
    # notice that: it returns the same values, it should be order-1 if no prediction is required BUT
    # we build the required values in a (hopefully) fast way, that requires
    # one point more.
    return ((order, None), (order, None))


def get_df(pv_array, suggested_step, predict=False):
    """The array must be built in this way:
    It has to be an array of arrays. Each of them has the following structure:

    [time, np_matrix, np_matrix]

    Hence the pv_array[k] element is made of:
    _ time is the time in which the solution is valid: t(n-k)
    _ The first np_matrix is x(n-k)
    _ The second is d(x(n-k))/dt
    Values that are not needed may be set to None and they will be disregarded.

    if predict == True, it needs one more point to give a prediction
    of x at the suggested step.

    Returns: None if the incorrect values were given, or quits.
    Otherwise returns an array:
    _ the [0] element is the np matrix of coeffiecients (Nx1) of x(n+1)
    _ the [1] element is the np matrix of constant terms (Nx1) of x(n+1)
    The derivative may be written as:
    d(x(n+1))/dt = ret[0]*x(n+1) + ret[1]"""

    if order is None:
        printing.print_general_error(
            "You must set Gear's order before using it! e.g. gear.order = 5")
        sys.exit(1)

    s = []
    s.append(0)
    for index in range(1, order + 2):
        s.append(suggested_step + pv_array[0][0] - pv_array[index - 1][0])

    # build e[k, i]
    e = np.zeros((order + 2, order + 2))
    for k_index in range(1, order + 2):
        for i_index in range(1, order + 2):
            if i_index == k_index:
                e[k_index, i_index] = 1
            else:
                e[k_index, i_index] = s[i_index] / (s[i_index] - s[k_index])

    alpha = np.zeros((1, order + 2))
    for k_index in range(1, order + 2):
        alpha[0, k_index] = 1.0
        for j_index in range(order + 1):
            alpha[0, k_index] = alpha[0, k_index] * e[k_index, j_index + 1]

    # build gamma
    gamma = np.zeros((1, order + 1))
    for k_index in range(1, order + 1):
        gamma[0, k_index] = alpha[0, k_index] * \
            ((1.0 / s[order + 1]) - (1.0 / s[k_index]))

    gamma[0, 0] = 0
    for index in range(1, order + 1):
        gamma[0, 0] = gamma[0, 0] - gamma[0, index]

    # values to be returned
    C1 = gamma[0, 0]

    C0 = np.zeros(pv_array[0][1].shape)
    for index in range(order):
        C0 = C0 + gamma[0, index + 1] * pv_array[index][1]

    x_lte_coeff = 0
    for k_index in range(1, order + 1):
        x_lte_coeff = x_lte_coeff + \
            (s[k_index] ** (order + 1)) * \
            (-1.0 * gamma[0, k_index] / gamma[0, 0])
    x_lte_coeff = ((-1.0) ** (order + 1)) * \
        (1.0/factorial(order + 1)) * x_lte_coeff

    if predict:
        predict_x = np.zeros(pv_array[0][1].shape)
        for index in range(1, order + 2):  # order
            predict_x = predict_x + alpha[0, index] * pv_array[index - 1][1]

        predict_lte_coeff = -1.0/factorial(order + 1)
        for index in range(1, order + 2):
            predict_lte_coeff = predict_lte_coeff * s[index]
        # print predict_lte_coeff
        # print x_lte_coeff
    else:
        predict_x = None
        predict_lte_coeff = None

    return C1, C0, x_lte_coeff, predict_x, predict_lte_coeff
